/* ::::::::::::::
 * includes.h
 * ::::::::::::::
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>
#include <assert.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "utils.h"
#include "config.h"

#define _ARGF_PRIVATE_H_

/* #include "argf.h" */
/* #include "argDesc.h" */
/* #include "argFind.h" */
/* #include "argPrint.h" */
/* #include "argActual.h" */
/* #include "argStore.h" */
/* #include "miscStr.h" */
/* #include "sizeOfReal.h" */

#include "argf.h"
/* ::::::::::::::
 * argDesc.h
 * ::::::::::::::
 */
extern int use_plus_minus_options;

#ifndef TRUE
#define TRUE		1
#define FALSE		0
#endif
#ifndef UNDEFINED
#define UNDEFINED	-1
#endif

typedef int boolean;
typedef int (*verifyfunc)(char*, char*, int);

enum ARG_JOIN {ARG_NOJOIN,ARG_OR,ARG_AND,ARG_SEQ,ARG_XOR};
enum ARG_TYPE {
  ARG_BAD_TYPE,
  ARG_LITERAL,
  ARG_INTEGER,
  ARG_FLOAT,
  ARG_REAL,
  ARG_DOUBLE,
  ARG_STRING,
  ARG_CHAR,
  ARG_BOOLEAN,
  ARG_NATURAL,
  ARG_SHORT,
  ARG_LONG,
  ARG_FORMAT,
  ARG_AREA_FORMAT,
  ARG_RANGE_STRING,
  ARG_VAR_NAME,
  ARG_MEMBER_NAME,
  ARG_FILE_IN,
  ARG_FILE_OUT
};

#define FOUND_ACTUAL		0x1
#define NO_ACTUAL		0x2
#define OPTION_ON		0x4
#define OPTION_OFF		0x8
#define OPTION_PRESENT		0x10
#define SATISFIED_ARGS		0x20
#define PARTIALLY_FILLED	0x40
#define FILLED_AND_CLOSED	0x80
#define ACTUAL_CLEARED		0x1000
#define ACTUAL_FLAGS (FOUND_ACTUAL|NO_ACTUAL|OPTION_ON|OPTION_OFF|OPTION_PRESENT|SATISFIED_ARGS|PARTIALLY_FILLED|FILLED_AND_CLOSED)
#define ALL_ACTUAL_FLAGS	0xFFFF

#define ARG_DESC_MAGIC_START	0x5c6b7a89
#define ARG_DESC_MAGIC_END	0xc5d4e3f2

struct ARG_DESC {
  int			magic_number;	/* ARG_MAGIC_START or ARG_MAGIC_END */
  struct ARG_LIST	**arg_list;
  int			n_lists;
  char			*name;
  int			filling_aps;	/* 1 if we are just filling aps, not
					   building the whole thing */
  int			fill_list;	/* next list to fill */
  int			complete;
  char			**option_names;	/* to check for ambiguities */
  int			(*func_ptr)();
};

struct ARG_LIST {
  struct ARG_GROUP	*group;
  enum ARG_JOIN		join;
  struct ARG_LIST	*next;
  struct ARG_LIST	*parent;
  int			id;		/* just to uniquely identify */
  void			***ap_loc;
  int			*ap_should_point_to_null;
  int			n_aps;
  int			actual;
};

struct ARG_GROUP {
  enum {ARG_GROUP_SIMPLE, ARG_GROUP_OPTION}
  				tag;
  union {
    struct {
      struct ARG_OPTION		*option;
      struct ARG_LIST		*arg_list;
      int			optional;
    }			option;
    struct {
      struct ARG_SIMPLE		*arg_simple;
    }			simple;
  }				u;
  int				*presence;
  int				*neg_presence;
  char				*help;
  int				actual;
};

struct ARG_SIMPLE {
  char			*prompt;
  int			vector;		/* argument is a vector */
  enum ARG_TYPE		type;
  verifyfunc		verify;
  void			*store;		/* where to store data */
  int			*store_len;	/* where to store vector length */
  int			vector_len;	/* where to store vector length */
  int			actual;		/* store info about actual arg */
  int			id;
};

struct ARG_OPTION {
  char			*name;
  int			no_prefix;
  int			is_signed;
  int			*sign;		/* store sign here */
  int			*neg_sign;	/* store !sign here */
  int			*presence;	/* store 1 here if present */
  int			*neg_presence;	/* store 0 here if present */
  char			*help;
  int			actual;		/* store info about actual arg */
};

/***
 * Different versions from argf.h that compile correctly with the
 * source, rather than specify correct usage (because of the varargs.h
 * stuff).
 */
extern void	Args(struct ARG_DESC *, char *, ...);
extern int	CountArgs(char **);
extern void	EndArgs(struct ARG_DESC *);
extern char **	ParseArgs(struct ARG_DESC *, int, char **, int);
extern struct ARG_DESC *StartArgs(char *);
/* ::::::::::::::
 * argFind.h
 * ::::::::::::::
 */
extern void	StoreArgDesc(struct ARG_DESC *);
extern int	GetPlusMinusOptionSwitch(void);
/* ::::::::::::::
 * argPrint.h
 * ::::::::::::::
 */
void PrintArgHelp(struct ARG_DESC *, FILE *file, int);
void PrintArgTree(struct ARG_DESC *, FILE *file);
void PrintArgUsage(struct ARG_DESC *, FILE *file, int);
void PrintArgValues(struct ARG_DESC *, FILE *file);

void PrintArgDesc(struct ARG_DESC *, int, int, FILE *);

extern char * SprintArgDesc(char*, struct ARG_DESC*, int, int);
extern char * SprintArgList(char*, struct ARG_LIST*, int, int);
extern char * SprintArgGroup(char*, struct ARG_GROUP*, int, int);
extern char * SprintArgOption(char*, struct ARG_OPTION*, int, int);
extern char * SprintArgSimple(char*, struct ARG_SIMPLE*, int, int);
extern char * SprintArgJoin(char*, enum ARG_JOIN, int, int);
extern char * SprintArgValue(char*, enum ARG_TYPE, void *, int);
extern char * InitBuf(char *);
extern char * FlagPrefix(int);

#define INSIDE_GROUP		0x1
#define PRINT_ADDRESSES		0x2
#define PRINT_BRIEF		0x4
#define PRINT_NESTED		0x8
#define PRINT_VALUES		0x10
#define PRINT_HELP		0x20
#define PRINT_PARENS		0x40
#define PRINT_BRACKETS		0x80
#define PRINT_PERCENTS		0x100
#define PRINT_SEQS		0x200
#define PRINT_ONEPERLINE	0x400
#define PRINT_ACTUALS		0x800
#define PRINT_SPECIFIERS	0x1000
#define PRINT_STRUCTS		0x2000
#define PRINT_TYPES		0x4000
#define PRINT_SIMPLES		0x8000
#define PRINT_LITERALS		0x10000
#define PRINT_OPTIONS		0x20000
#define PRINT_OPTION_DASH	0x40000
#define PRINT_NOT_INDENT	0x80000
#define PRINT_OPTION_SIGN	0x100000
#define PRINT_NOT_NEWLINE	0x200000
#define PRINT_ACTUAL_STATUS	0x400000
#define PRINT_GROUPPERLINE	0x800000
#define	PRINT_OBJECTS (PRINT_LITERALS|PRINT_SIMPLES|PRINT_OPTIONS|PRINT_OPTION_DASH)
#define PRINT_BRIEF_LINE (PRINT_OBJECTS|PRINT_BRACKETS|PRINT_NOT_NEWLINE)
/* ::::::::::::::
 * argActual.h
 * ::::::::::::::
 */
struct GLOBAL {
		/* to know whether when we can continue vector filling */
  int			last_simple_id;
  struct ARG_DESC	*arg_desc;
  int			debug;
  int			usage_on_error;
};

extern char **	ParseArgs(struct ARG_DESC *, int, char **, int);
extern void	ArgErrorAbort(struct GLOBAL *, char *, ...);
extern void	ArgError(struct GLOBAL *, char *, ...);
extern int	ArgNElements(char **);

/* ::::::::::::::
 * argStore.h
 * ::::::::::::::
 */
extern boolean argBoolean(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argString(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argChar(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argNatural(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argInteger(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argShort(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argLong(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argReal(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argFloat(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argDouble(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argFormatStr(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argAreaFormatStr(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argRangeStr(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argVarName(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argMemberName(struct ARG_SIMPLE *, char *, int, struct GLOBAL*);
extern boolean argFileIn(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern boolean argFileOut(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);
extern void	autoinit_argtypes(void);

#define NUMBER_ARG_TYPES	20
#define COPYSTR		0x1
#define	ONLYCHECK	0x2
#define	OPTIONAL	0x4

typedef int (*ArgStoreFunction)(struct ARG_SIMPLE *, char *, int, struct GLOBAL *);

extern enum ARG_TYPE arg_type_by_char[128];
extern ArgStoreFunction arg_store_function_by_type[NUMBER_ARG_TYPES];
extern char* arg_type_name[NUMBER_ARG_TYPES];
extern char arg_type_char[NUMBER_ARG_TYPES];
extern int arg_types_initialized;
/* ::::::::::::::
 * miscStr.h
 * ::::::::::::::
 */
extern char *pchar(char);
extern void strncpy_null(char*, char*, int);
extern char *strcpyend(char*, char*);
extern char *strcatend(char*, char*);

/* ::::::::::::::
 * sizeOfReal.h
 * ::::::::::::::
 */
void ArgsSetRealSize(int size);
int doubleReals(void);
/* ::::::::::::::
 * argDesc.c
 * ::::::::::::::
 */
/* #include "includes.h" */

/***
 * This file parses argument descriptions, and builds the structures
 * corresponding to them.
 */

struct GLOBALS {
  va_list	ap;
  void	***ap_loc;	/* where the varargs got stored */
  int		*ap_should_point_to_null;
  int		i;		/* next vararg to use */
  int		N;		/* total number of varargs */
  char		*spec;		/* the original line */
};

#define SKIPSPACE(str) while (isspace(*(str))) (str)++

static int ArgList
 (char*, char**, struct ARG_LIST**, struct GLOBALS *, struct ARG_LIST *);
static int ArgJoin
 (char*, char**, enum ARG_JOIN*, enum ARG_JOIN);
static int ArgGroup
 (char*, char**, struct ARG_GROUP**, struct GLOBALS *, struct ARG_LIST *);
static int ArgOption
 (char*, char**, struct ARG_OPTION**, struct GLOBALS *);
static int ArgSimple
 (char*, char**, struct ARG_SIMPLE**, struct GLOBALS *);
static int ArgStore
 (char*, char**, struct ARG_SIMPLE*, struct GLOBALS *);

extern struct ARG_DESC *FindOldArgDesc(char*);

struct ARG_DESC *StartArgs
  (char *name)
{
  struct ARG_DESC *arg_desc;
  if (!arg_types_initialized)
    autoinit_argtypes();
  if (name==NULL) {
    fprintf(stderr,"Args: Warning: call to StartArgs() has supplied no name\n");
    name = "(nil)";
  }
  if (!IsAsciiString(name))
    fprintf(stderr,"Args: Warning: call to StartArgs() has given non-ascii name\n");
  arg_desc = FindOldArgDesc(name);
  if (arg_desc==NULL) {
    arg_desc = CallocOrAbort(struct ARG_DESC, 1);
    arg_desc->arg_list = NULL;
    arg_desc->filling_aps = 0;
    arg_desc->name = CallocOrAbort(char, strlen(name)+1);
    arg_desc->fill_list = -1;
    arg_desc->n_lists = 0;
    arg_desc->option_names = NULL;
    strcpy(arg_desc->name,name);
  } else {
    arg_desc->filling_aps = 1;
    arg_desc->fill_list = 0;
  }
  arg_desc->magic_number = ARG_DESC_MAGIC_START;
  arg_desc->complete = 0;
  return arg_desc;
}

void Args
  (struct ARG_DESC	*arg_desc,
   char			*arg_spec,
   ...)
{
  
  struct ARG_LIST *arg_list;
  int i;
  va_list ap;
  va_start(ap, arg_spec);
  if (arg_desc->magic_number!=ARG_DESC_MAGIC_START) {
    if (arg_desc->magic_number==ARG_DESC_MAGIC_END)
      ErrorAbort("Args: bad magic number for argd record, EndArgs() already called?");
    else
      ErrorAbort("Args: bad magic number for argd record");
  }
  if (arg_desc->filling_aps) {
    /***
     * fill in the arg pointer slots with the var args passed to use
     */
    if (arg_desc->fill_list<0 || arg_desc->fill_list>=arg_desc->n_lists)
      ErrorAbort("Args: no more fill lists for spec \"%s\"",arg_spec);
    arg_list = arg_desc->arg_list[arg_desc->fill_list];
    for (i=0;i<arg_list->n_aps;i++) {
      *(arg_list->ap_loc[i]) = va_arg(ap, void*);
      if (   *(void**)(arg_list->ap_loc[i])==NULL
	  || *(void**)(arg_list->ap_loc[i])==(void*)1) {
	Error("Args: argument %d (after arg spec) is not a pointer", i+1);
	ErrorAbort("Command \"%s\", arg specification: \"%s\"",
		    arg_desc->name, arg_spec);
      }
      if (   arg_list->ap_should_point_to_null[i]
	  && **(void***)(arg_list->ap_loc[i])!=NULL) {
	Error("Args: vector pointer argument %d (after arg spec) should point to NULL", i+1);
	ErrorAbort("Command \"%s\", arg specification: \"%s\"",
		    arg_desc->name, arg_spec);
      }
    }
    arg_desc->fill_list++;
  } else {
    /***
     * parse the string and build the arg tree
     */
    struct GLOBALS globals;
    char *end;
    globals.ap = ap;
    globals.N = 0;
    globals.i = 0;
    globals.spec = arg_spec;
    end = arg_spec;
    /* count the number of % parameters */
    while ((end=index(end,'%'))) {
      globals.N++;
      end++;
    }
    globals.N++; /* allow for a NULL at the end */
    globals.ap_loc = CallocOrAbort(void **, globals.N);
    globals.ap_should_point_to_null = CallocOrAbort(int, globals.N);
    for (i=0;i<globals.N;i++) {
      globals.ap_loc[i] = NULL;
      globals.ap_should_point_to_null[i] = 0;
    }
    if (!ArgList(arg_spec, &end, &arg_list, &globals, NULL))
      ErrorAbort("Args: Failed to parse \"%s\"",arg_spec);
    SKIPSPACE(end);
    if (*end!='\0')
      ErrorAbort("Args: extra junk \"%s\" at end of \"%s\"", end, arg_spec);
    if (globals.i >= globals.N)
      ErrorAbort("Args: used too many aps in parsing \"%s\"", arg_spec);
    arg_list->n_aps = globals.i;
    arg_list->ap_loc = globals.ap_loc;
    arg_list->ap_should_point_to_null = globals.ap_should_point_to_null;
    if (arg_desc->arg_list==NULL) {
      arg_desc->arg_list = CallocOrAbort(struct ARG_LIST*, 2);
      arg_desc->arg_list[0] = arg_list;
      arg_desc->n_lists = 1;
    } else {
      arg_desc->arg_list
	= ReallocOrAbort(struct ARG_LIST*, arg_desc->arg_list,
			 arg_desc->n_lists+2);
      arg_desc->arg_list[arg_desc->n_lists] = arg_list;
      arg_desc->n_lists++;
    }
    arg_desc->arg_list[arg_desc->n_lists] = NULL;
    for (i=0;i<arg_list->n_aps;i++) {
      if (   *(void**)(arg_list->ap_loc[i])==NULL
	  || *(void**)(arg_list->ap_loc[i])==(void*)1) {
	Error("Args: argument %d (after arg spec) is not a pointer", i+1);
	ErrorAbort("Command \"%s\", arg specification: \"%s\"",
		   arg_desc->name, arg_spec);
      }
    }
  }
  va_end(ap);
}

void EndArgs
  (struct ARG_DESC *arg_desc)
{
  if (arg_desc->magic_number!=ARG_DESC_MAGIC_START) {
    if (arg_desc->magic_number==ARG_DESC_MAGIC_END)
      ErrorAbort("EndArgs: bad magic number for argd record, EndArgs() already called?");
    else
      ErrorAbort("EndArgs: bad magic number for argd record");
  }

  if (arg_desc->filling_aps && arg_desc->fill_list!=arg_desc->n_lists)
    ErrorAbort("EndArgs: was filling aps, but fill_list!=n_lists");

  if (arg_desc->option_names==NULL) {
    /***
     * Build a list of the option names
     */
#define ENDARGSBUFSIZ 1000
    char buffer[ENDARGSBUFSIZ];/* big enough to hold all option names! */
    char *end, *start;
    char **name;
    int n = 0;		/* count the number of options */
    end = SprintArgDesc(InitBuf(buffer), arg_desc,
			PRINT_OPTIONS|PRINT_NOT_INDENT|PRINT_ONEPERLINE, 0);
    assert(end < buffer+ENDARGSBUFSIZ);
    end = index(buffer+1, '\n');
    while (end!=NULL) {
      end++;
      n++;
      end = index(end, '\n');
    }
    name = arg_desc->option_names = CallocOrAbort(char *, n+1);
    start = buffer+1;
    while (1) {
      end = index(start, '\n');
      if (end==NULL)
	break;
      assert(end-start > 0);
      *name = CallocOrAbort(char, end-start+1);
      strncpy_null(*name, start, end-start+1);
      start = end+1;
      name++;
    }
    *name = NULL;
  }
  arg_desc->filling_aps = 0;
  arg_desc->complete = 1;
  arg_desc->magic_number = ARG_DESC_MAGIC_END;
  StoreArgDesc(arg_desc);
}

/***
 * ArgList
 * - Parses:
ARGLIST -> [ ARGGROUP ARGMORE ]
ARGLIST -> ARGGROUP ARGNEXT
ARGLIST -> NULL

ARGNEXT	-> ARGJOIN ARGLIST			- default JOIN is AND
ARGNEXT	-> NULL

ARGMORE	-> ARGJOIN ARGLIST			- default JOIN is AND
ARGMORE	-> NULL
 */
static int ArgList
  (char *str,
   char **end_out,
   struct ARG_LIST **retobj,
   struct GLOBALS *globals,
   struct ARG_LIST *parent)
{
  struct ARG_LIST *arg_list;
  struct ARG_LIST *top_arg_list = NULL;
  struct ARG_LIST *tail_arg_list = NULL;
  struct ARG_GROUP *arg_group;
  static int arg_list_id = 0;
  enum ARG_JOIN join;
  char *end;
  int found_join = 0;

  while (1) {
    SKIPSPACE(str);
    arg_list = CallocOrAbort(struct ARG_LIST, 1);
    arg_list->id = ++arg_list_id;
    if (ArgGroup(str, &end, &arg_group, globals, arg_list)) {
      str = end;
      arg_list->group = arg_group;
      arg_list->next = NULL;
      arg_list->join = ARG_NOJOIN;
      arg_list->parent = parent;
      found_join = ArgJoin(str, &end, &join, ARG_SEQ);
      str = end;
      if (top_arg_list!=NULL)
	tail_arg_list->next = arg_list;
      else
	top_arg_list = arg_list;
      arg_list->join = join;
      tail_arg_list = arg_list;
    } else {
      free(arg_list);
      arg_list_id--;
      break;
    }
  }
  if (found_join)
    ErrorAbort("expected arg, found \"%s\"\nIn: \"%s\"",
		str, globals->spec);

  if (top_arg_list==NULL)
    return 0;
  else {
    *end_out = str;
    *retobj = top_arg_list;
    return 1;
  }
}

/***
 * ArgGroup
 * - Parses:
ARGGROUP-> ( ARGBODY PRESENCE HELP )	- () is optional

ARGBODY	-> SIMPLEARG
ARGBODY	-> OPTION ARGLIST
ARGBODY	-> [ OPTION ARGLIST ]
ARGBODY -> [ ARGLIST ]

PRESENCE-> NULL
PRESENCE-> % [!] P

HELP	-> NULL
HELP	-> %?
 */

static int ArgGroup
  (char *str,
   char **end_out,
   struct ARG_GROUP **retobj,
   struct GLOBALS *globals,
   struct ARG_LIST *parent)
{
  struct ARG_GROUP *arg_group;
  struct ARG_OPTION *arg_option = NULL;
  struct ARG_SIMPLE *arg_simple = NULL;
  struct ARG_LIST *arg_list = NULL;
  int looking_for_paren = 0;	/* around the ARGGROUP */
  int have_option = 0;
  int have_arg_list = 0;
  int have_arg_simple = 0;
  int optional = 0;
  int stop;
  char *marker = NULL;

  SKIPSPACE(str);

  if (*str=='\0')
    return 0;

  if (*str=='(') {
    looking_for_paren = 1;
    str++;
  }

  SKIPSPACE(str);
  if (*str=='[') {
    marker = ++str;
    optional = 1;
  }
  have_option = ArgOption(str, &str, &arg_option, globals);
  if (have_option || optional)
    have_arg_list = ArgList(str, &str, &arg_list, globals, parent);
  else
    have_arg_simple = ArgSimple(str, &str, &arg_simple, globals);

  if (optional && !have_option && !have_arg_list)
    ErrorAbort("Args: expected arg list after \"%s\", but found \"%s\"\nIn: \"%s\"",
		marker, str, globals->spec);

  if (optional) {
    SKIPSPACE(str);
    if (*str!=']')
      ErrorAbort("Args: expected closing \"]\", but found \"%s\"\nIn: \"%s\"", str, globals->spec);
    str++;
  }

  if (!have_option && !have_arg_simple && !have_arg_list) {
    if (looking_for_paren)
      ErrorAbort("Args: cannot parse \"%s\"\nIn: \"%s\"", str, globals->spec);
    else
      return 0;
  }

  arg_group = CallocOrAbort(struct ARG_GROUP, 1);
  if (optional || have_option) {
    arg_group->tag = ARG_GROUP_OPTION;
    arg_group->u.option.option = arg_option;
    arg_group->u.option.arg_list = arg_list;
    arg_group->u.option.optional = optional;
  } else {
    arg_group->tag = ARG_GROUP_SIMPLE;
    arg_group->u.simple.arg_simple = arg_simple;
  }
  arg_group->presence = NULL;
  arg_group->neg_presence = NULL;
  arg_group->help = NULL;

  str = str;
  SKIPSPACE(str);

  stop = 0;
  while (*str=='%' && !stop) {	/* find %P and %? */
    if (str[1]=='P' && arg_group->presence==NULL) {
      arg_group->presence = va_arg(globals->ap, int*);
      globals->ap_loc[globals->i++] = (void**) &arg_group->presence;
      str += 2;
    } else if (str[1]=='!' && str[2]=='P' && arg_group->neg_presence==NULL) {
      arg_group->neg_presence = va_arg(globals->ap, int*);
      globals->ap_loc[globals->i++] = (void**) &arg_group->neg_presence;
      str += 3;
    } else if (str[1]=='?' && arg_group->help==NULL) {
      arg_group->help = va_arg(globals->ap, char*);
      globals->ap_loc[globals->i++] = (void**) &arg_group->help;
      str += 2;
    } else {
      stop = 1;
      break;
    }
    if (!stop)
      SKIPSPACE(str);
  }

  SKIPSPACE(str);
  if (looking_for_paren) {
    if (*str!=')')
      ErrorAbort
	("Args: expected closing \")\" for arg group, but found \"%s\"\nIn: \"%s\"", str, globals->spec);
    str++;	/* skip the ) */
  }
  
  *end_out = str;
  *retobj = arg_group;

  return 1;
    
}

/***
 * ArgSimple
 * - Parses:
SIMPLEARG	-> <description> ARGVECT {ARGSTORE}
SIMPLEARG	-> "string"
SIMPLEARG	-> 'string'

ARGVECT -> NULL
ARGVECT	-> ...
 */

static int ArgSimple
  (char *str,
   char **end_out,
   struct ARG_SIMPLE **retobj,
   struct GLOBALS *globals)
{
  char *end, *start;
  struct ARG_SIMPLE *arg_simple;
  static int arg_simple_id = 1;
  start = str;
  SKIPSPACE(str);
  if (*str=='<') {
    end = str;
    while (*end && *end!='>') end++;			/* find the <  > */
    if (*end=='\0')
      ErrorAbort("Args: no terminating '>' for string \"%s\"\nIn: \"%s\"",
		  str, globals->spec);

    arg_simple = CallocOrAbort(struct ARG_SIMPLE, 1);	/* allocate */
    arg_simple->prompt = CallocOrAbort(char, end-str);
    arg_simple->id = arg_simple_id++;
    strncpy_null(arg_simple->prompt, str+1, end-str);

    str = end+1;
    SKIPSPACE(str);					/* look for a "..." */
    if (str[0]=='.' && str[1]=='.' && str[2]=='.') {
      arg_simple->vector = 1;
      str += 3;
    } else
      arg_simple->vector = 0;

    if (!ArgStore(str, &end, arg_simple, globals))
      ErrorAbort
	("Args: expected arg store spec (something like \"%i\"), but found \"%s\"\nIn: \"%s\"",
	 str, globals->spec);
    str = end;
  } else if (*str=='\'' || *str=='"') {
    end = str+1;
    while (*end && *end!=*str) end++;
    if (*end=='\0')
      ErrorAbort("Args: no terminating %c for string \"%s\"\nIn: \"%s\"",
		  *str, str, globals->spec);
    arg_simple = CallocOrAbort(struct ARG_SIMPLE, 1);
    arg_simple->id = arg_simple_id++;
    arg_simple->prompt = CallocOrAbort(char, end-str);
    strncpy_null(arg_simple->prompt, str+1, end-str);
    arg_simple->vector = 0;
    arg_simple->type = ARG_LITERAL;
    arg_simple->verify = NULL;
    arg_simple->store = NULL;
    end++;
    str = end;
  } else
    return 0;

  *end_out = end;
  *retobj = arg_simple;
  return 1;

}

/***
 * ArgStore
 * - Parses: {ARGSTORE} (at least one)
ARGSTORE	-> % ARGTYPE

ARGTYPE	-> V				- verify function provided
ARGTYPE	-> L				- length of vector

ARGTYPE	-> i				- integer
ARGTYPE	-> f				- float
ARGTYPE	-> r				- real
ARGTYPE	-> D				- double
ARGTYPE	-> s				- string
ARGTYPE	-> c				- char
ARGTYPE	-> B				- boolean
ARGTYPE	-> n				- natural number (integer >= 0)
ARGTYPE	-> h				- short
ARGTYPE	-> l				- long
ARGTYPE	-> F				- format (as in %9.3g etc.)
ARGTYPE	-> A				- area format
ARGTYPE	-> R				- range string
ARGTYPE	-> v				- variable name
ARGTYPE	-> m				- member (field) name
 */

static int ArgStore
  (char *str,
   char **end_out,
   struct ARG_SIMPLE *arg_simple,
   struct GLOBALS *globals)
{
  char *end;
  end = str;
  SKIPSPACE(end);
  if (*end!='%') return 0;
  str = end;
  while (*str=='%') {
    str++;
    if (*str=='V') {
      arg_simple->verify = va_arg(globals->ap, verifyfunc);
      globals->ap_loc[globals->i] = (void**) &arg_simple->verify;
    } else if (*str=='L') {
      if (!arg_simple->vector)
	ErrorAbort("Args: <%s> is not a vector argument - cannot use '%%L'\nIn: \"%s\"",
		   arg_simple->prompt, globals->spec);
      arg_simple->store_len = va_arg(globals->ap, int*);
      globals->ap_loc[globals->i] = (void**) &arg_simple->store_len;
    } else if (arg_type_by_char[(int) *str]!=ARG_BAD_TYPE) {
      arg_simple->type = arg_type_by_char[(int) *str];
      arg_simple->store = va_arg(globals->ap, void*);
      globals->ap_loc[globals->i] = &arg_simple->store;
      if (arg_simple->vector) {
	globals->ap_should_point_to_null[globals->i] = 1;
	if (*(char**)arg_simple->store!=NULL)
	  ErrorAbort("Args: pointer for arg list <%s>... contains non-null value\nIn: \"%s\"",
		     arg_simple->prompt, globals->spec);
      }
    } else {
      /* must be a %P or %? that gets parsed elsewhere */
      break;
    }
    globals->i++;
    end = ++str;
    SKIPSPACE(str);
  }
  *end_out = end;
  return 1;
}

/***
 * ArgJoin
 * - Parses:
ARGJOIN -> ^ XOR				- one but not both
ARGJOIN -> -> SEQ				- both in this sequence
ARGJOIN -> & AND				- both in any sequence
ARGJOIN -> | OR 				- one or both (or neither)
ARGJOIN -> NULL (default, OR or AND)
 */
static int ArgJoin
  (char *str,
   char **end_out,
   enum ARG_JOIN *retobj,
   enum ARG_JOIN default_join)
{
  char *end;
  end = str;
  SKIPSPACE(end);
  if (*end=='^') {
    end++;
    *retobj = ARG_XOR;
  } else if (*end=='+') {
    end++;
    *retobj = ARG_SEQ;
  } else if (*end=='&') {
    end++;
    *retobj = ARG_AND;
  } else if (*end=='|') {
    end++;
    *retobj = ARG_OR;
  } else {
    *retobj = default_join;
    return 0;
  }
  *end_out = end;
  return 1;
}

/***
 * ArgOption
 * - Parses:
OPTION	-> '-' OPTFLAGS OPTTYPE OPTNAME OPTSTORE HELP
 */
static int ArgOption
  (char *str,
   char **end_out,
   struct ARG_OPTION **retobj,
   struct GLOBALS *globals)
{
  struct ARG_OPTION *arg_option;
  char *mark = NULL;
  int stop = 0;
  char *start = str;
  SKIPSPACE(str);
  if (*str!='-') {					/* not an option */
    return 0;
  } else {						/* is an option */
    str++;						/* skip the - */
    arg_option = CallocOrAbort(struct ARG_OPTION, 1);
    arg_option->no_prefix = 0;
    while (isalpha(*str) && !(*str=='S' || *str=='P')) {/* flags */
      switch (*str) {
      case 'F':
	arg_option->no_prefix = 1;
	break;
      default:
	ErrorAbort("Args: bad flag %s in option \"%s\"\nIn: \"%s\"", pchar(*str), start, globals->spec);
      }
      str++;
    }
    if (*str=='S')					/* type */
      arg_option->is_signed = 1;
    else if (*str=='P')
      arg_option->is_signed = 0;
    else
      ErrorAbort("Args: can't find type of option ('B' or 'P') in \"%s\"\nIn: \"%s\"", start, globals->spec);
    str++;

    SKIPSPACE(str);

    mark = str;
    while (isalnum(*str) || *str=='-') str++;			/* name */
    arg_option->name = CallocOrAbort(char, 1+str-mark);
    strncpy_null(arg_option->name, mark, 1+str-mark);

    SKIPSPACE(str);
    while (*str=='%' && !stop) {
      if (str[1]=='!') {
	switch (str[2]) {
	case 'S' :
	  arg_option->neg_sign = va_arg(globals->ap, int*);
	  globals->ap_loc[globals->i++] = (void**) &arg_option->neg_sign;
	  break;
	case 'P' :
	  arg_option->neg_presence = va_arg(globals->ap, int*);
	  globals->ap_loc[globals->i++] = (void**)&arg_option->neg_presence;
	  break;
	default:
	  stop = 1;
	  break;
	}
	if (!stop) {
	  str += 3;
	  SKIPSPACE(str);
	}
      } else {	/* not negated */
	switch (str[1]) {
	case 'S' :
	  arg_option->sign = va_arg(globals->ap, int*);
	  globals->ap_loc[globals->i++] = (void**) &arg_option->sign;
	  break;
	case 'P' :
	  arg_option->presence = va_arg(globals->ap, int*);
	  globals->ap_loc[globals->i++] = (void**) &arg_option->presence;
	  break;
	case '?' :
	  arg_option->help = va_arg(globals->ap, char*);
	  globals->ap_loc[globals->i++] = (void**) &arg_option->help;
	  break;
	default:
	  stop = 1;
	  break;
	}
	if (!stop) {
	  str += 2;
	  SKIPSPACE(str);
	}
      }
    }
    *end_out = str;
    *retobj = arg_option;
    return 1;
  }
}
/* ::::::::::::::
 * argActual.c
 * ::::::::::::::
 */
/* #include "includes.h" */

/***
 * parseact.c - parse the actual arguments
 * To fill an arglist (A):
 *   find an unused arg_list (B) that an actual (X) fits in
 *   put it there, if (X) is an option, then clear out the arglist (B)
 *   fill arglist (B)
 *   fill the parent arglist (C), if it is not = (A)
 *   mark (A) as used
 *   check if (A) is satisfied (all non-optional args filled),
 *     if not print error message
 *
 * For the filling of vector arguments, we keep record of where
 * we used the last actual argument, then we can add to a vector arg
 * if we put the last actual in the same place
 */

int use_plus_minus_options = 0;

enum SET_MODE {
  SET_ON,
  SET_OFF};

enum OPTION_FILL_RESULT {
  NOT_FILLED,
  FILLED_FIRST_TIME,
  FILLED_AGAIN};

enum FIND_GROUP_RESULT {
  NOT_FOUND_NOT_FULL,
  NOT_FOUND_EMPTY,
  NOT_FOUND_FULL,
  FOUND_GROUP_OPTION_OR_ARGLIST,
  FOUND_GROUP_SIMPLE};

static int IsOption(char *);

static char **ParseArgsDesc(struct ARG_DESC*, char**, int, int, int, int);

static struct ARG_LIST *FindActualInArgDesc
 (struct ARG_DESC*, char*, struct GLOBAL *);
static struct ARG_LIST *FindActualInArgList
 (struct ARG_LIST*, char*, struct GLOBAL *);
static enum FIND_GROUP_RESULT FindActualInArgGroup
 (struct ARG_GROUP*, char*, struct GLOBAL*);

static char **FillArgList(struct ARG_LIST*, char**, struct GLOBAL*);
static int FillArgSimple
 (struct ARG_SIMPLE*, char*, struct GLOBAL*);
static enum OPTION_FILL_RESULT FillArgOption
 (struct ARG_OPTION*, char*, struct GLOBAL*);

static void CheckArgListSatisfied
 (struct ARG_LIST*, char**, struct GLOBAL*);
static void CheckArgGroupSatisfied
 (struct ARG_GROUP*, char**, struct GLOBAL*);

static void SAFArgDesc(struct ARG_DESC*, int, enum SET_MODE);
static void SAFArgList(struct ARG_LIST*, int, enum SET_MODE);
static void SAFArgGroup(struct ARG_GROUP*, int, enum SET_MODE);
static void SAFArgSimple(struct ARG_SIMPLE*, int, enum SET_MODE);
static void SAFArgOption(struct ARG_OPTION*, int, enum SET_MODE);

char **ParseArgs
  (struct ARG_DESC *arg_desc,
   int argc,
   char *argv[],
   int flags)
{
  struct ARG_DESC *my_arg_desc;
  int arg_help = 0;
  int arg_values = 0;
  int arg_usage = 0;
  int arg_debug = 0;
  int arg_tree = 0;
  if (argc<0)
    return argv;
  if (arg_desc->magic_number!=ARG_DESC_MAGIC_END)
    ErrorAbort
      ("ParseArgs: bad magic number for argd \"%s\".\nWrong argument order or has not been processed by EndArgs()?",
       arg_desc->name);
  my_arg_desc = StartArgs("argargs");
  use_plus_minus_options = GetPlusMinusOptionSwitch();
  Args(my_arg_desc, "[-FSargplusminus %S]", &use_plus_minus_options);
  Args(my_arg_desc, "[-FPargtree %P]", &arg_tree);
  Args(my_arg_desc, "[-FPargdebug %P]", &arg_debug);
  Args(my_arg_desc, "[-FPargusage %P]", &arg_usage);
  Args(my_arg_desc, "[-FPargvalues %P]", &arg_values);
  Args(my_arg_desc, "[-FParghelp %P]", &arg_help);
  EndArgs(my_arg_desc);
  argv = ParseArgsDesc(my_arg_desc, argv+1, 1, 1, 0, 0);
  if (arg_debug) {
    fprintf(dout, "Argument debugging turned on, internal argument flags:\n");
    fprintf(dout, "  arg_help=%s arg_values=%s arg_usage=%s arg_debug=%s\n",
	    BooleanString(arg_help), BooleanString(arg_values),
	    BooleanString(arg_usage), BooleanString(arg_debug));
    fprintf(dout, "  arg_tree=%s use_plus_minus_options=%s\n",
	    BooleanString(arg_tree), BooleanString(use_plus_minus_options));
  }
  if (arg_help)
    PrintArgHelp(arg_desc, dout, 1);
  else if (arg_usage)
    PrintArgUsage(arg_desc, dout, 1);
  else {
    argv = ParseArgsDesc(arg_desc, argv, (flags&EXTRA_ARGS)!=0,
			 (flags&EXTRA_OPTIONS)!=0, 1, arg_debug);
    if (!(arg_values || arg_tree || arg_debug))
      return argv;				/* normal return */
    else {
      if (arg_values)
	PrintArgValues(arg_desc, dout);
      if (arg_tree)
	PrintArgTree(arg_desc, dout);
    }
  }
  if (!(flags&RETURN_ON_ERROR)) {
    Abort();
    return NULL;
  } else
    return NULL;	/* indication to stop! */
}

static char **ParseArgsDesc
  (struct ARG_DESC *arg_desc,
   char *argv[],
   int extra_args,
   int extra_opts,
   int insert_name,
   int debug)
{
  int i;
  struct ARG_LIST *arg_list;
  struct GLOBAL globals;
  char **extra;
  i=0;
  while (argv[i]!=0) i++;
  extra = ITempCalloc(char*, i+2);
  i = 0;
  if (insert_name)
    extra[i++] = arg_desc->name;
  globals.arg_desc = arg_desc;
  globals.last_simple_id = 0;
  globals.debug = debug;
  globals.usage_on_error = 1;
  SAFArgDesc(arg_desc, ALL_ACTUAL_FLAGS, SET_OFF);
  while (*argv) {
    arg_list = FindActualInArgDesc(arg_desc, *argv, &globals);
    if (arg_list)
      argv = FillArgList(arg_list, argv+1, &globals);
    else if (IsOption(*argv))
      if (extra_opts) {
	extra[i++] = *argv++;
	globals.last_simple_id = 0;
      } else
	ArgErrorAbort(&globals, "option \"%s\" not understood",*argv);
    else if (extra_args) {
      extra[i++] = *argv++;
      globals.last_simple_id = 0;
    } else
      ArgErrorAbort(&globals, "argument \"%s\" not understood",*argv);
  }
  extra[i] = NULL;
  if (arg_desc->arg_list!=NULL)
    for (i=0; i<arg_desc->n_lists; i++)
      CheckArgListSatisfied(arg_desc->arg_list[i], extra, &globals);
  return extra;
}

/***
 * this function does two things - try to fill up the arg_list
 * and also checks that it is satisfied
 */
static char **FillArgList
  (struct ARG_LIST *arg_list,
   char **argv,
   struct GLOBAL *global)
{
  struct ARG_LIST *sub_arg_list = NULL;
  if (*argv!=NULL) {
    sub_arg_list = FindActualInArgList(arg_list, *argv, global);
    if (sub_arg_list==NULL) {
      CheckArgListSatisfied(arg_list, argv, global);
      return argv;
    }
    argv++;
  }
  if (*argv!=NULL)
    argv = FillArgList(sub_arg_list, argv, global);
  if (*argv!=NULL && sub_arg_list!=arg_list)
    argv = FillArgList(arg_list, argv, global);
  if (*argv!=NULL && arg_list->parent!=NULL)
    argv = FillArgList(arg_list->parent, argv, global);
  CheckArgListSatisfied(arg_list, argv, global);
  return argv;
}

static void CheckArgListSatisfied
  (struct ARG_LIST *arg_list,
   char **argv,
   struct GLOBAL *global)
{
  if (arg_list->actual&SATISFIED_ARGS)
    return;
  if (arg_list->join==ARG_XOR && (arg_list->next->actual&PARTIALLY_FILLED))
    CheckArgListSatisfied(arg_list->next, argv, global);
  else {
    /***
     * the following stuff is an attempt to do args that are `joined'
     * with OR, AND, SEQ or XOR, but it doesn't work (or maybe the
     * join values are not being read properly).
     */
    if (   (arg_list->group->actual&PARTIALLY_FILLED)
	|| arg_list->join==ARG_SEQ || arg_list->join==ARG_AND
	|| arg_list->join==ARG_XOR || arg_list->join==ARG_NOJOIN)
      CheckArgGroupSatisfied(arg_list->group, argv, global);
    if (   arg_list->next!=NULL
	&& (   (arg_list->next->actual&PARTIALLY_FILLED)
	    || arg_list->join==ARG_SEQ || arg_list->join==ARG_AND))
      CheckArgListSatisfied(arg_list->next, argv, global);
    if (   arg_list->join==ARG_OR
	&& !(arg_list->group->actual&SATISFIED_ARGS)
	&& !(arg_list->next->actual&SATISFIED_ARGS)) {
      char g_buf[1000];
      char l_buf[1000];
      (void) SprintArgGroup(InitBuf(g_buf), arg_list->group,
			    PRINT_BRIEF_LINE, 0);
      (void) SprintArgList(InitBuf(l_buf), arg_list->next,
			   PRINT_BRIEF_LINE, 0);
      ArgErrorAbort(global, "expected argument to put in \"%s\" or \"%s\"",
		  g_buf+1, l_buf+1);
    }
  }
  arg_list->actual |= SATISFIED_ARGS;
}

static void CheckArgGroupSatisfied
  (struct ARG_GROUP *arg_group,
   char **argv,
   struct GLOBAL *global)
{
  if (arg_group->actual&SATISFIED_ARGS)
    return;
  if (arg_group->tag==ARG_GROUP_SIMPLE) {
    if (!(arg_group->u.simple.arg_simple->actual&FOUND_ACTUAL)) {
      char g_buf[1000];
      (void) SprintArgGroup(InitBuf(g_buf), arg_group, PRINT_BRIEF_LINE, 0);
      if (g_buf[1]=='\'') /* is ARG_LITERAL */
	ArgErrorAbort(global, "Expected argument to match %s", g_buf+1);
      else
	ArgErrorAbort(global, "Expected argument to put in \"%s\"", g_buf+1);
    }
    if (arg_group->presence)
      *(arg_group->presence) = 1;
    if (arg_group->neg_presence)
      *(arg_group->neg_presence) = 0;
  } else {
    struct ARG_LIST *list = arg_group->u.option.arg_list;
    struct ARG_OPTION *option = arg_group->u.option.option;
    assert(arg_group->tag==ARG_GROUP_OPTION);
    /***
     * the group can be satisfied by it being optional
     * & having nothing, or by having an option & the arg_list
     * being satisfied
     */
    if (   arg_group->u.option.optional
	&& (option==NULL || !(option->actual&FOUND_ACTUAL))
	&& (list==NULL || !(list->actual&PARTIALLY_FILLED))) {
      /* it's ok - optional & have nothing */
    } else {
      if (option!=NULL && !(option->actual&FOUND_ACTUAL)) {
	char o_buf[1000];
	(void) SprintArgOption(InitBuf(o_buf), option, PRINT_BRIEF_LINE, 0);
	ArgErrorAbort(global, "Expected option \"%s\"", o_buf+1);
      }
      if (list!=NULL) {
	CheckArgListSatisfied(list, argv, global);
	assert(list->actual&SATISFIED_ARGS);
      }
      if (arg_group->presence)
	*(arg_group->presence) = 1;
      if (arg_group->neg_presence)
	*(arg_group->neg_presence) = 0;
    }
  }
  arg_group->actual |= SATISFIED_ARGS;
}

/***
 * this fills an argument
 */
static int FillArgSimple
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   struct GLOBAL *global)
{
  /***
   * we only try to fill if either
   * (1) the formal arg is unfilled, or
   * (2) the formal arg is a vector type, and the last actual was used here
   */
  if (!(   !(arg_simple->actual&FOUND_ACTUAL)
	|| (arg_simple->vector && global->last_simple_id==arg_simple->id))) {
    return 0;
  } else {
    /*
     * to fill:
     * (1) check if it is of the right type,
     * (2) check if it passes the verify function
     * (3) store the argument
     */
    int right_type = 0;
    int passes_verify = 1;
    if (arg_simple->type==ARG_LITERAL) {	/* literal should just match */
      if (arg_simple->verify==NULL)
	right_type = !strcmp(arg, arg_simple->prompt);
      else
	right_type = 1;				/* unless a verify function */
    } else {					/* check the type */
      right_type = arg_store_function_by_type[(int) arg_simple->type]
	(arg_simple, arg, ONLYCHECK, global);
    }
    if (global->debug)
      fprintf(dout, "Argument \"%s\" %s type check with %c%s%c (%s)\n",
	      arg, (right_type ? "passed" : "failed"),
	      (arg_simple->type==ARG_LITERAL ? '\'' : '<'),
	      arg_simple->prompt,
	      (arg_simple->type==ARG_LITERAL ? '\'' : '>'),
	      arg_type_name[(int) arg_simple->type]);
    if (!right_type)				/* not the right type */
      return 0;
    if (arg_simple->verify!=NULL) {		/* use verify function */
      passes_verify = arg_simple->verify
	(arg, arg_simple->prompt, global->debug);
      if (global->debug)
	fprintf(dout, "Argument \"%s\" %s verify check with %c%s%c (%s)\n",
		arg, (passes_verify ? "passed" : "failed"),
		(arg_simple->type==ARG_LITERAL ? '"' : '<'),
		arg_simple->prompt,
		(arg_simple->type==ARG_LITERAL ? '"' : '>'),
		arg_type_name[(int) arg_simple->type]);
    }
    if (!passes_verify)				/* didn't pass verify */
      return 0;
    if (arg_simple->type!=ARG_LITERAL) {	/* don't store literals */
      (void) arg_store_function_by_type[(int)arg_simple->type]
	(arg_simple, arg, COPYSTR, global);
    }
    arg_simple->actual |= FOUND_ACTUAL;
    global->last_simple_id = arg_simple->id;
    return 1;
  }
}

static enum OPTION_FILL_RESULT FillArgOption
  (struct ARG_OPTION *arg_option,
   char *arg,
   struct GLOBAL *global)
{
  enum OPTION_FILL_RESULT retval;
  int pos_match = 0;
  int pos_perfect_match = 0;
  int other_pos_match = 0;
  int other_pos_perfect_match = 0;
  int other_pos_match_n = 0;
  int neg_match = 0;
  int neg_perfect_match = 0;
  int other_neg_match = 0;
  int other_neg_perfect_match = 0;
  int other_neg_match_n = 0;
  int i, len, match;
  char *p, *pos_arg, *neg_arg;
  char *pos_flag, *neg_flag, *ambiguity;
  char **name = global->arg_desc->option_names;
  char debug_intro[100];
  if (!IsOption(arg))
    return NOT_FILLED;
  assert(*arg=='-' || *arg=='+');
  debug_intro[0] = '\0';

  /* strip off the '-' or '+' and look for a prefix match */
  p = arg+1;
  pos_arg = p;
  len = strlen(p);
  match = !strncasecmp(p, arg_option->name, len);
  pos_perfect_match = match && (len == strlen(arg_option->name));
  pos_match = (match && (!arg_option->no_prefix || pos_perfect_match));
  if (global->debug && match) {
    sprintf(debug_intro, "Matching \"%s\" with target option \"%s\"\n",
	    arg, arg_option->name);
    if (pos_perfect_match)
      fprintf(dout, "%s  Perfect match of \"[%c]%s\" with \"%s\"\n",
	      debug_intro, *arg, p, arg_option->name);
    else if (match)
      fprintf(dout, "%s  Matched prefix of \"[%c]%s\" with \"%s\"%s\n",
	      debug_intro, *arg, p, arg_option->name, !pos_match ?
	      ", but partial matches are not accepted for this option" : "");
  }

  /* strip off the "no" and look for a prefix match */
  if (arg[0]=='-' && p[0]=='n' && p[1]=='o' && len>2) {
    p += 2;
    len -= 2;
    neg_arg = p;
    match = !strncasecmp(p, arg_option->name, len);
    neg_perfect_match = match && (len == strlen(arg_option->name));
    neg_match = (match && (!arg_option->no_prefix || neg_perfect_match));
    if (global->debug && match) {
      if (debug_intro[0]=='\0')
	sprintf(debug_intro, "Matching \"%s\" with target option \"%s\"\n",
		arg, arg_option->name);
      else
	debug_intro[0]='\0';
      if (neg_perfect_match)
	fprintf(dout, "%s  Perfect match of \"[-no]%s\" with \"%s\"\n",
		debug_intro, p, arg_option->name);
      else if (match)
	fprintf(dout, "%s  Matched prefix of \"[-no]%s\" with \"%s\"%s\n",
		debug_intro, p, arg_option->name, !neg_match ?
		", but partial matches are not accepted for this option" : "");
    }
  } else
    neg_arg = NULL;

  if (!pos_match && !neg_match)
    return NOT_FILLED;

  /* look for ambiguities */
  if (pos_match || neg_match) {
    len = strlen(pos_arg);
    for (i=0;name[i]!=NULL;i++) {
      if (   strcasecmp(arg_option->name, name[i])
	  && !strncasecmp(pos_arg, name[i], len)
	  && (!pos_perfect_match || len==strlen(name[i]))) {
	other_pos_match_n = i;
	other_pos_match = 1;
	other_pos_perfect_match = (len == strlen(name[i]));
	if (global->debug)
	  fprintf(dout, "  %s of \"[%c]%s\" with \"%s\"\n",
		  (other_pos_perfect_match
		   ? "Perfect match" : "Matched prefix"),
		   *arg, pos_arg, name[i]);
	break;
      }
    }

    if (neg_arg) {
      len = strlen(neg_arg);
      for (i=0;name[i]!=NULL;i++) {
	if (   strcasecmp(arg_option->name, name[i])
	    && !strncasecmp(neg_arg, name[i], len)
	    && (!neg_perfect_match || len==strlen(name[i]))) {
	  other_neg_match_n = i;
	  other_neg_match = 1;
	  other_neg_perfect_match = (len == strlen(name[i]));
	  if (global->debug)
	    fprintf(dout, "  %s of \"[-no]%s\" with \"%s\"\n",
		    (other_neg_perfect_match
		     ? "Perfect match" : "Matched prefix"),
		    neg_arg, name[i]);
	  break;
	}
      }
    }
  }

  assert(!(pos_perfect_match && neg_perfect_match));
  if (  pos_perfect_match + neg_perfect_match
      + other_pos_perfect_match + other_neg_perfect_match > 1) {
    ArgError(global, "Command options are inherently ambiguous\n\
The source should be changed, ambiguous options are: ");
    if (pos_perfect_match)
      ArgError(global, "%s", arg);
    if (neg_perfect_match)
      ArgError(global, "-no%s", arg+1);
    if (other_pos_perfect_match)
      ArgError(global, "%c%s", *arg, name[other_pos_match_n]);
    if (other_neg_perfect_match)
      ArgError(global, "-no%s", name[other_neg_match_n]);
    Abort();
  }
  /***
   * perfect matches override, so if we matched other perfectly,
   * but not this one, return 
   */
  if (   (other_neg_perfect_match || other_pos_perfect_match)
      && !(pos_perfect_match || neg_perfect_match)) {
    if (global->debug)
      fprintf(dout, "  Perfect match with other option overrides partial match with target option\n");
    return NOT_FILLED;
  }
  /***
   * and if we matched ours perfectly, then ignore others
   */
  if (pos_perfect_match || neg_perfect_match) {
    if (global->debug && (other_pos_match || other_neg_match))
      fprintf(dout, "  Perfect match with target option overrides partial match with other option\n");
    other_pos_match = other_neg_match = 0;
  }
  if (pos_perfect_match) {
    if (global->debug && neg_match)
      fprintf(dout, "  Perfect match with target option overrides partial match with [-no]target\n");
    neg_match = 0;
  }
  if (neg_perfect_match) {
    if (global->debug && pos_match)
      fprintf(dout, "  Perfect match with [-no]target option overrides partial match with target\n");
    pos_match = 0;
  }
  /***
   * now, check for any remaining ambiguity
   */
  pos_flag = (use_plus_minus_options ? "+" : "-");
  neg_flag = (use_plus_minus_options ? "-" : "-no");
  ambiguity = "Ambigous option \"%s\", could mean \"%s%s\" or \"%s%s\"";
  if (pos_match && (neg_match || other_neg_match))
    ArgErrorAbort(global, ambiguity, arg, neg_flag, arg_option->name, neg_flag,
		(neg_match ? arg_option->name : name[other_neg_match_n]));
  if (neg_match && (pos_match || other_pos_match))
    ArgErrorAbort(global, ambiguity, arg, neg_flag, arg_option->name, pos_flag,
		(pos_match ? arg_option->name : name[other_neg_match_n]));
  if (pos_match && other_pos_match)
    ArgErrorAbort(global, ambiguity, arg, pos_flag, arg_option->name, pos_flag,
		name[other_pos_match_n]);
  if (neg_match && other_neg_match)
    ArgErrorAbort(global, ambiguity, arg, neg_flag, arg_option->name, neg_flag,
		name[other_neg_match_n]);

  /* should be no ambiguity now */
  assert(pos_match + neg_match + other_pos_match + other_neg_match == 1);
  
  /* convert to a real pos/neg - taking into account use_plus_minus_options */
  if (use_plus_minus_options && *arg=='-') {
    pos_match = 0;
    neg_match = 1;
  }
  if (neg_match && !arg_option->is_signed)
    ArgErrorAbort(global, "option \"%s%s\" is not signed - cannot use \"%s%s\"",
		pos_flag, arg_option->name, neg_flag, arg_option->name);

  if (arg_option->actual && FOUND_ACTUAL)
    retval = FILLED_AGAIN;
  else
    retval = FILLED_FIRST_TIME;
  arg_option->actual =
    (FOUND_ACTUAL | (pos_match
		     ? (arg_option->is_signed ? OPTION_ON : OPTION_PRESENT)
		     : OPTION_OFF));
  if (arg_option->presence!=NULL) {
    *(arg_option->presence) = 1;
    if (global->debug)
      fprintf(dout, "  Presence of \"%s\" set to true\n", arg_option->name);
  }
  if (arg_option->neg_presence!=NULL) {
    *(arg_option->neg_presence) = 0;
    if (global->debug)
      fprintf(dout, "  Neg presence of \"%s\" set to false\n",
	      arg_option->name);
  }
  if (arg_option->sign!=NULL) {
    *(arg_option->sign) = pos_match;
    if (global->debug)
      fprintf(dout, "  Sign of \"%s\" set to %s\n", arg_option->name,
	      BooleanString(pos_match));
  }
  if (arg_option->neg_sign!=NULL) {
    *(arg_option->neg_sign) = !pos_match;
    if (global->debug)
      fprintf(dout, "  Neg sign of \"%s\" set to %s\n", arg_option->name,
	      BooleanString(!pos_match));
  }
  global->last_simple_id = 0;	/* not any simple_id */
  return retval;
}

static struct ARG_LIST *FindActualInArgDesc
  (struct ARG_DESC *arg_desc,
   char *arg,
   struct GLOBAL *global)
{
  int i;
  struct ARG_LIST *arg_list;
  for (i=0; i<arg_desc->n_lists; i++) {
    arg_list = FindActualInArgList(arg_desc->arg_list[i], arg, global);
    if (arg_list!=NULL)
      return arg_list;
  }
  return NULL;
}

static struct ARG_LIST *FindActualInArgList
  (struct ARG_LIST *arg_list,
   char *arg,
   struct GLOBAL *global)
{
  enum FIND_GROUP_RESULT find_group_result = NOT_FOUND_EMPTY;
  int no_do_group = 0;
  struct ARG_LIST *next;

  next = arg_list->next;
  /***
   * have to do some sequencing stuff here
   * we try to fill the group first, (unless join=XOR && arglist is partial)
   *
   * retval from FindActualInArgGroup
   * 			/ join	OR/AND		XOR		SEQ
   * NOT_FOUND_NOT_FULL		Try next	Ret NULL	Ret NULL
   * NOT_FOUND_EMPTY		Try next	Try next	Ret NULL
   * NOT_FOUND_FULL		Try next	Ret NULL	Try next
   * FOUND_GROUP_OPTION_OR_ARG_LIST -- Ret group_arg_list (arg_list if NULL) --
   * FOUND_GROUP_SIMPLE_ARG	    -----     Ret arg_list    -----
   */
  no_do_group = (   arg_list->group==NULL
		 || (   next!=NULL && arg_list->join==ARG_XOR
		     && (next->actual & PARTIALLY_FILLED)));
  if (!no_do_group)
    find_group_result = FindActualInArgGroup(arg_list->group, arg, global);
  else
    find_group_result = NOT_FOUND_EMPTY;

  if (   find_group_result==FOUND_GROUP_SIMPLE
      || find_group_result==FOUND_GROUP_OPTION_OR_ARGLIST)
    arg_list->actual |= PARTIALLY_FILLED;
  
  if (find_group_result==FOUND_GROUP_SIMPLE)
    return arg_list;

  if (find_group_result==FOUND_GROUP_OPTION_OR_ARGLIST) {
    if (arg_list->group->u.option.arg_list!=NULL)
      return arg_list->group->u.option.arg_list;
    else
      return arg_list;
  }

  if (next==NULL)
    return NULL;

  if (   (arg_list->join==ARG_AND || arg_list->join==ARG_OR)
      || (arg_list->join==ARG_XOR && find_group_result==NOT_FOUND_EMPTY)
      || (arg_list->join==ARG_SEQ && find_group_result==NOT_FOUND_FULL))
      return FindActualInArgList(arg_list->next, arg, global);
  else
    return NULL;
}

static enum FIND_GROUP_RESULT FindActualInArgGroup
  (struct ARG_GROUP *arg_group,
   char *arg,
   struct GLOBAL *global)
{
  if (arg_group->tag==ARG_GROUP_SIMPLE) {
    int complete_but_space =
      (   arg_group->u.simple.arg_simple!=NULL
       && (arg_group->u.simple.arg_simple->actual & FOUND_ACTUAL)
       && arg_group->u.simple.arg_simple->vector
       && global->last_simple_id==arg_group->u.simple.arg_simple->id);
    
    if (arg_group->u.simple.arg_simple==NULL
	|| (   (arg_group->u.simple.arg_simple->actual & FOUND_ACTUAL)
	    && !complete_but_space))
      return NOT_FOUND_FULL;
    else if (   !IsOption(arg)
	     && FillArgSimple(arg_group->u.simple.arg_simple, arg, global)) {
      arg_group->actual |= PARTIALLY_FILLED;
      return FOUND_GROUP_SIMPLE;
    } else {
      if (complete_but_space)
	return NOT_FOUND_FULL;
      else
	return NOT_FOUND_EMPTY;
    }
  } else {
    enum OPTION_FILL_RESULT opt_fill_res;
    struct ARG_LIST *arg_list;
    assert( arg_group->tag==ARG_GROUP_OPTION );
    /***
     * if arg is an option && option is not NULL, try to fill the option
     * otherwise, try to fill the arglist if it is ready
     */
    if (IsOption(arg) && arg_group->u.option.option!=NULL) {
      opt_fill_res = FillArgOption(arg_group->u.option.option, arg, global);
      if (opt_fill_res==FILLED_AGAIN && arg_group->u.option.arg_list!=NULL) {
	SAFArgList(arg_group->u.option.arg_list, ACTUAL_CLEARED, SET_ON);
	SAFArgList(arg_group->u.option.arg_list, ACTUAL_FLAGS, SET_OFF);
	arg_group->actual |= ACTUAL_CLEARED;	/* set this */
	arg_group->actual &= ~ACTUAL_FLAGS;	/* clear these */
	if (global->debug) {
	  char buf[1000];
	  (void) SprintArgGroup(InitBuf(buf), arg_group, PRINT_BRIEF_LINE, 0);
	  fprintf(dout,"  Option \"%s\" found again, clearing sublist \"%s\"\n",
		  arg_group->u.option.option->name, buf+1);
	}
      }
      if (opt_fill_res==FILLED_FIRST_TIME || opt_fill_res==FILLED_AGAIN) {
	arg_group->actual = PARTIALLY_FILLED;
	return FOUND_GROUP_OPTION_OR_ARGLIST;
      }
    }

    /***
     * return if the option is not filled
     */
    if (   arg_group->u.option.option!=NULL
	&& !(arg_group->u.option.option->actual & FOUND_ACTUAL)) {
      if (   arg_group->u.option.arg_list!=NULL
	  && !(arg_group->u.option.arg_list->actual & PARTIALLY_FILLED))
	return NOT_FOUND_NOT_FULL;
      else
	return NOT_FOUND_EMPTY;
    }

    /***
     * if there is an option, it is full.
     * now, try to fill the arg_list
     */
    if (   arg_group->u.option.arg_list==NULL
	|| (arg_group->u.option.arg_list->actual&FILLED_AND_CLOSED))
      return NOT_FOUND_FULL;

    arg_list = FindActualInArgList(arg_group->u.option.arg_list, arg, global);

    if (arg_list==NULL) {	/* didn't find */
      if (arg_group->u.option.arg_list->actual & SATISFIED_ARGS) {
	arg_group->u.option.arg_list->actual |= FILLED_AND_CLOSED;
	return NOT_FOUND_FULL;
      } else {
	return NOT_FOUND_NOT_FULL;
      }
    } else {
      return FOUND_GROUP_OPTION_OR_ARGLIST;
    }
  }
  /* gcc might complain here, but other compilers can work out that this
     line can't be reached */
}


/***
 * SAF* functions "Set All Flags"
 */
static void SAFArgDesc
  (struct ARG_DESC *arg_desc,
   int flags,
   enum SET_MODE mode)
{
  int i;
  for (i=0; i<arg_desc->n_lists; i++)
    SAFArgList(arg_desc->arg_list[i], flags, mode);
}

static void SAFArgList
  (struct ARG_LIST *arg_list,
   int flags,
   enum SET_MODE mode)
{
  if (arg_list->group!=NULL)
    SAFArgGroup(arg_list->group, flags, mode);
  if (arg_list->next!=NULL)
    SAFArgList(arg_list->next, flags, mode);
  if (mode==SET_ON)
    arg_list->actual |= flags;
  else if (mode==SET_OFF)
    arg_list->actual &= ~flags;
  else
    assert(0);
}

static void SAFArgGroup
  (struct ARG_GROUP *arg_group,
   int flags,
   enum SET_MODE mode)
{
  if (arg_group->tag==ARG_GROUP_OPTION) {
    if (arg_group->u.option.option!=NULL)
      SAFArgOption(arg_group->u.option.option, flags, mode);
    if (arg_group->u.option.arg_list!=NULL)
      SAFArgList(arg_group->u.option.arg_list, flags, mode);
  } else if (arg_group->tag==ARG_GROUP_SIMPLE) {
    if (arg_group->u.simple.arg_simple!=NULL)
      SAFArgSimple(arg_group->u.simple.arg_simple, flags, mode);
  }
  if (mode==SET_ON)
    arg_group->actual |= flags;
  else if (mode==SET_OFF)
    arg_group->actual &= ~flags;
  else
    assert(0);
}

static void SAFArgSimple
  (struct ARG_SIMPLE *simple_arg,
   int flags,
   enum SET_MODE mode)
{
  if (mode==SET_ON)
    simple_arg->actual |= flags;
  else if (mode==SET_OFF)
    simple_arg->actual &= ~flags;
  else
    assert(0);
}

static void SAFArgOption
  (struct ARG_OPTION *arg_option,
   int flags,
   enum SET_MODE mode)
{
  if (mode==SET_ON)
    arg_option->actual |= flags;
  else if (mode==SET_OFF)
    arg_option->actual &= ~flags;
  else
    assert(0);
}

static int IsOption
  (char *arg)
{
  if ((*arg=='-' || *arg=='+') && isalpha(arg[1])) return 1;
  else return 0;
}

void ArgErrorAbort
  (struct GLOBAL	*global,
   char			*message,
   ...)
{
  va_list ap;

  va_start(ap, message);
  ErrorV(message, &ap);
  if (global->usage_on_error) {
    PrintArgUsage(global->arg_desc, stderr, 1);
  }
  Abort();
}

void ArgError
  (struct GLOBAL	*global,
   char			*message,
   ...)
{
  va_list ap;

  va_start(ap, message);
  ErrorV(message, &ap);
}

int ArgNElements
  (char **ptr)
{
  if (ptr==NULL)
    return 0;
  else {
    int i = 0;
    for (; *ptr!=NULL; ptr++)
      i++;
    return i;
  }
}

int CountArgs(char *argv[])
{
  int i = 0;
  if (argv!=NULL)
    while (argv[i]!=NULL)
      i++;
  return i;
}
/* ::::::::::::::
 * argStore.c
 * ::::::::::::::
 */
/* #include "includes.h" */
/* #include <sys/types.h> */
/* #include <sys/stat.h> */
/* #include <errno.h> */

/*
extern long strtol(char*, char**, int);
extern double strtod(char*, char**);
*/

/*
**  ARGtype -- argument translation routines.
**
**	Each of these converts a parameter value to the internal form,
**	including validity checking.  Their parameters and return values
**	all behave identically.
**
**	Parameters:
**		arg_simple -- the argument descriptor for this parameter.
**		arg -- a pointer to the string input value.
**		flags - COPYSTR
**			-- the value will be destroyed later,
**			and so should be copied if it will be retained
**			(as for a string).
**			OPTIONAL
**			-- no error messages are given on failure
**			CHECK
**			-- nothing at all is changed, and TRUE is
**			returned if the value can be converted
**			
**
**	Returns:
**		TRUE -- if the conversion was successful.  The actual
**			value should be stored in the location indicated
**			by arg_simple->store.
**		FALSE -- if the conversion failed.  If optional==FALSE,
**			an error message should be posted with Error()
**
**	Side Effects:
**		The value should be stored through arg_simple->store.
*/

enum ARG_TYPE arg_type_by_char[128];
ArgStoreFunction arg_store_function_by_type[NUMBER_ARG_TYPES];
char* arg_type_name[NUMBER_ARG_TYPES];
char arg_type_char[NUMBER_ARG_TYPES];
int is_string_type[NUMBER_ARG_TYPES];
int arg_types_initialized = 0;

void autoinit_argtypes(void)
{
  int i;
  arg_types_initialized = 1;
  for (i=0;i<128;i++)
    arg_type_by_char[i] = ARG_BAD_TYPE;
  for (i=0;i<NUMBER_ARG_TYPES;i++) {
    arg_store_function_by_type[i] = NULL;
    arg_type_name[i] = "bad type";
    arg_type_char[i] = '\0';
  }
  arg_type_by_char['i'] = ARG_INTEGER;
  arg_type_by_char['f'] = ARG_FLOAT;
  arg_type_by_char['r'] = ARG_REAL;
  arg_type_by_char['D'] = ARG_DOUBLE;
  arg_type_by_char['s'] = ARG_STRING;
  arg_type_by_char['c'] = ARG_CHAR;
  arg_type_by_char['B'] = ARG_BOOLEAN;
  arg_type_by_char['n'] = ARG_NATURAL;
  arg_type_by_char['h'] = ARG_SHORT;
  arg_type_by_char['l'] = ARG_LONG;
  arg_type_by_char['F'] = ARG_FORMAT;
  arg_type_by_char['A'] = ARG_AREA_FORMAT;
  arg_type_by_char['R'] = ARG_RANGE_STRING;
  arg_type_by_char['v'] = ARG_VAR_NAME;
  arg_type_by_char['m'] = ARG_MEMBER_NAME;
  arg_type_by_char['I'] = ARG_FILE_IN;
  arg_type_by_char['O'] = ARG_FILE_OUT;
  for (i=0;i<128;i++)
    if (arg_type_by_char[i]!=ARG_BAD_TYPE)
      arg_type_char[(int) arg_type_by_char[i]] = i;

  arg_type_name[(int)ARG_LITERAL] = "literal";
  arg_type_name[(int)ARG_INTEGER] = "integer";
  arg_type_name[(int)ARG_FLOAT] = "float";
  arg_type_name[(int)ARG_REAL] = "real";
  arg_type_name[(int)ARG_DOUBLE] = "double";
  arg_type_name[(int)ARG_STRING] = "string";
  arg_type_name[(int)ARG_CHAR] = "char";
  arg_type_name[(int)ARG_BOOLEAN] = "boolean";
  arg_type_name[(int)ARG_NATURAL] = "natural";
  arg_type_name[(int)ARG_SHORT] = "short";
  arg_type_name[(int)ARG_LONG] = "long";
  arg_type_name[(int)ARG_FORMAT] = "format";
  arg_type_name[(int)ARG_AREA_FORMAT] = "area format";
  arg_type_name[(int)ARG_RANGE_STRING] = "range string";
  arg_type_name[(int)ARG_VAR_NAME] = "var name";
  arg_type_name[(int)ARG_MEMBER_NAME] = "member name";
  arg_type_name[(int)ARG_FILE_IN] = "input file";
  arg_type_name[(int)ARG_FILE_OUT] = "output file";

  arg_store_function_by_type[(int)ARG_INTEGER] = argInteger;
  arg_store_function_by_type[(int)ARG_FLOAT] = argFloat;
  arg_store_function_by_type[(int)ARG_REAL] = argReal;
  arg_store_function_by_type[(int)ARG_DOUBLE] = argDouble;
  arg_store_function_by_type[(int)ARG_STRING] = argString;
  arg_store_function_by_type[(int)ARG_CHAR] = argChar;
  arg_store_function_by_type[(int)ARG_BOOLEAN] = argBoolean;
  arg_store_function_by_type[(int)ARG_NATURAL] = argNatural;
  arg_store_function_by_type[(int)ARG_SHORT] = argShort;
  arg_store_function_by_type[(int)ARG_LONG] = argLong;
  arg_store_function_by_type[(int)ARG_FORMAT] = argFormatStr;
  arg_store_function_by_type[(int)ARG_AREA_FORMAT] = argAreaFormatStr;
  arg_store_function_by_type[(int)ARG_RANGE_STRING] = argRangeStr;
  arg_store_function_by_type[(int)ARG_VAR_NAME] = argVarName;
  arg_store_function_by_type[(int)ARG_MEMBER_NAME] = argMemberName;
  arg_store_function_by_type[(int)ARG_FILE_IN] = argFileIn;
  arg_store_function_by_type[(int)ARG_FILE_OUT] = argFileOut;

  is_string_type[(int)ARG_INTEGER] = 0;
  is_string_type[(int)ARG_FLOAT] = 0;
  is_string_type[(int)ARG_REAL] = 0;
  is_string_type[(int)ARG_DOUBLE] = 0;
  is_string_type[(int)ARG_STRING] = 1;
  is_string_type[(int)ARG_CHAR] = 0;
  is_string_type[(int)ARG_BOOLEAN] = 0;
  is_string_type[(int)ARG_NATURAL] = 0;
  is_string_type[(int)ARG_SHORT] = 0;
  is_string_type[(int)ARG_LONG] = 0;
  is_string_type[(int)ARG_FORMAT] = 1;
  is_string_type[(int)ARG_AREA_FORMAT] = 1;
  is_string_type[(int)ARG_RANGE_STRING] = 1;
  is_string_type[(int)ARG_VAR_NAME] = 1;
  is_string_type[(int)ARG_MEMBER_NAME] = 1;
  is_string_type[(int)ARG_FILE_IN] = 0;
  is_string_type[(int)ARG_FILE_OUT] = 0;
}

/***
 * Macro to add a value into a simple arg - three cases have to be
 * handled - ordinary scalar, vector of string types,
 * and vector of other types.
 */
#define ADD_VALUE(ARG_SIMPLE, TYPE, VAL) \
  if (!ARG_SIMPLE->vector) {\
    bcopy((char*)&VAL, (char*)(ARG_SIMPLE->store), sizeof(TYPE));\
  } else if (is_string_type[(int)ARG_SIMPLE->type]) {\
    TYPE **argv = *(TYPE ***) ARG_SIMPLE->store;\
    TYPE *cp;\
    int n;\
    if (argv!=NULL) {\
      n = ARG_SIMPLE->vector_len;\
      argv = ITempRealloc(TYPE *, argv, n+2);\
    } else {\
      n = ARG_SIMPLE->vector_len = 0;\
      argv = ITempCalloc(TYPE *, 2);\
    }\
    cp = ITempCalloc(TYPE, 1);\
    bcopy((char*)&(VAL), (char*)cp, sizeof(TYPE));\
    argv[n] = cp;\
    argv[n+1] = NULL;\
    *(TYPE ***) ARG_SIMPLE->store = argv;\
    ARG_SIMPLE->vector_len++;\
    if (ARG_SIMPLE->store_len!=NULL)\
      *ARG_SIMPLE->store_len = ARG_SIMPLE->vector_len;\
  } else {\
    TYPE *argv = *(TYPE **) ARG_SIMPLE->store;\
    int n;\
    if (argv!=NULL) {\
      n = ARG_SIMPLE->vector_len;\
      argv = ITempRealloc(TYPE , argv, n+2);\
    } else {\
      n = ARG_SIMPLE->vector_len = 0;\
      argv = ITempCalloc(TYPE , 2);\
    }\
    bcopy((char*)&(VAL), (char*)(argv+n), sizeof(TYPE));\
    bzero((char*)(argv+n+1), sizeof(TYPE));\
    *(TYPE **) ARG_SIMPLE->store = argv;\
    ARG_SIMPLE->vector_len++;\
    if (ARG_SIMPLE->store_len!=NULL)\
      *ARG_SIMPLE->store_len = ARG_SIMPLE->vector_len;\
  }

ArgStoreFunction arg_store_function[] =
{
  NULL,
  NULL,
  argInteger,
  argFloat,
  argReal,
  argDouble,
  argString,
  argChar,
  argBoolean,
  argNatural,
  argShort,
  argLong,
  argFormatStr,
  argAreaFormatStr,
  argRangeStr,
  argVarName,
  argMemberName
};

boolean
StoreArg(arg_simple, arg, flags, global)
struct ARG_SIMPLE *arg_simple;
char *arg;
int flags;
struct GLOBAL *global;
{
  if (   (int)(arg_simple->type) < 0
      || (int)(arg_simple->type) > (int) ARG_MEMBER_NAME
      || arg_store_function[(int)arg_simple->type] == NULL)
    ArgErrorAbort(global, "bad type %d in simple arg\n");
  return
    arg_store_function[(int)arg_simple->type](arg_simple, arg, flags, global);
}

boolean argChar
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  char *argp;
  int status = FALSE;
  char c;
  
  if (strlen(arg) == 2 && arg[0]=='^') {
    c = arg[1] ^ '@';
    status = TRUE;
  } else if (strlen(arg) > 1 && arg[0]=='\\') {
    c = (int) strtol(&arg[1], &argp, 8);
    if (*argp == '\0')
      status = TRUE;
  } else if (strlen(arg) == 1) {
    c = *arg;
    status = TRUE;
  }
  
  if (status == TRUE) {
    if (!(flags&ONLYCHECK)) {
      ADD_VALUE(arg_simple, char, c);
    }
  } else if (!(flags&(OPTIONAL|ONLYCHECK))) {
    ArgError(global, "invalid character argument '%s' for <%s>", arg, arg_simple->prompt);
    return FALSE;
  }
  return status;
}

boolean argString
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  char *cp;
  
  if (flags&ONLYCHECK)
    return TRUE;

  if (flags&COPYSTR) {
    int i;
    i = strlen(arg) + 1;
    cp = ITempCalloc(char, i);
    strcpy(cp, arg);
  } else {
    cp = arg;
  }

  if (arg_simple->vector) {
    char **argv = *(char ***) arg_simple->store;
    int n;
    if (argv==NULL) {
      argv = ITempCalloc(char *, 2);
      arg_simple->vector_len = n = 0;
    } else {
      n = arg_simple->vector_len;
      argv = ITempRealloc(char *, argv, n+2);
    }
    argv[n] = cp;
    argv[n+1] = NULL;
    *(char ***) arg_simple->store = argv;
    arg_simple->vector_len++;
    if (arg_simple->store_len!=NULL)
      *arg_simple->store_len = arg_simple->vector_len;
  } else
    *(char **) arg_simple->store = cp;

  return (TRUE);
}

boolean argRangeStr
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  int is_range=1;
  char *p = arg;
  /* go through a simple fsm to recognize a range */
  if (*p=='\0')
    is_range=0;
  else if (*p=='X')
    p++;
  else {
    while (isdigit(*p)) p++;
    if (*p=='-') p++;
    while (isdigit(*p)) p++;
    if (*p=='+') p++;
    while (isdigit(*p)) p++;
  }
  if (*p!='\0')
    is_range = 0;
  if (!is_range) {
    if (!(flags&(OPTIONAL|ONLYCHECK)))
      ArgError(global, "invalid range argument \"%s\" for <%s>", arg, arg_simple->prompt);
    return FALSE;
  }
  return argString(arg_simple, arg, flags, global);
}

boolean argFormatStr
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  int count=0;
  char *p = arg;
  char prev = '\0';
  /* go through the string & count the active %'s */
  for (;*p;p++) {
    if (*p=='%') {
      if (prev=='%') {
	count--;
	prev = '\0';
	continue;
      } else if (prev!='\\')
	count++;
    }
    prev = *p;
  }
  if (count!=1) {
    if (!(flags&(OPTIONAL|ONLYCHECK)))
      ArgError(global, "invalid print format argument \"%s\" for <%s>",
	       arg, arg_simple->prompt);
    return FALSE;
  }
  return argString(arg_simple, arg, flags, global);
}

boolean argMemberName
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  if (arg[0]=='.' || (arg[0]=='-' && arg[1]=='>'))
    return argString(arg_simple, arg, flags, global);
  else {
    if (!(flags&(OPTIONAL|ONLYCHECK)))
      ArgError(global, "invalid field name argument \"%s\" for <%s>",
	       arg, arg_simple->prompt);
    return FALSE;
  }
}

boolean argVarName
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  if (IsLegalVarName(arg))
    return argString(arg_simple, arg, flags, global);
  else {
    if (!(flags&(OPTIONAL|ONLYCHECK)))
      ArgError(global, "invalid field name argument \"%s\" for <%s>",
	       arg, arg_simple->prompt);
    return FALSE;
  }
}

boolean argAreaFormatStr
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  char *p = arg;
  int good = 0;
  if (isdigit(*p)) {
    while (isdigit(*p)) p++;
    if (*p=='x') {
      p++;
      if (isdigit(*p)) {
	while (isdigit(*p)) p++;
	if (*p=='j') p++;
	if (*p=='\0')
	  good = 1;
      }
    }
  }
  if (good)
    return argString(arg_simple, arg, flags, global);
  else {
    if (!(flags&(OPTIONAL|ONLYCHECK)))
      ArgError(global, "invalid area format argument \"%s\" for <%s>",
	       arg, arg_simple->prompt);
    return FALSE;
  }
}

/*ARGSUSED*/
boolean argInteger
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  char *argp;
  int value;
  
  value = strtol(arg, &argp, 0);
  if (*argp != '\0') {
    if (!(flags&(OPTIONAL|ONLYCHECK)))
      ArgError(global, "invalid integer argument \"%s\" for <%s>", arg, arg_simple->prompt);
    return (FALSE);
  } else {
    if (!(flags&ONLYCHECK)) {
      ADD_VALUE(arg_simple, int, value);
    }
    /* *(int *) arg_simple->store = value; */
    return (TRUE);
  }
}

/*ARGSUSED*/
boolean argNatural
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  char *argp;
  int value;
  
  value = strtol(arg, &argp, 0);
  if (*argp != '\0' || value<0) {
    if (!(flags&(OPTIONAL|ONLYCHECK)))
      ArgError(global, "invalid natural number argument \"%s\" for <%s>",
		  arg, arg_simple->prompt);
    return (FALSE);
  } else {
    if (!(flags&ONLYCHECK)) {
      ADD_VALUE(arg_simple, int, value);
    }
    /* *(int *) arg_simple->store = value; */
    return (TRUE);
  }
}

/*ARGSUSED*/
boolean argShort
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  char *argp;
  short value;
  
  value = strtol(arg, &argp, 0);
  if (*argp != '\0') {
    if (!(flags&(OPTIONAL|ONLYCHECK)))
      ArgError(global, "invalid integer argument \"%s\" for <%s>", arg, arg_simple->prompt);
    return (FALSE);
  } else {
    if (!(flags&ONLYCHECK)) {
      ADD_VALUE(arg_simple, short, value);
    }
    /* *(short *) arg_simple->store = value; */
    return (TRUE);
  }
}

/*ARGSUSED*/
boolean argLong
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  char *argp;
  long value;
  
  value = strtol(arg, &argp, 0);
  if (*argp != '\0') {
    if (!(flags&(OPTIONAL|ONLYCHECK)))
      ArgError(global, "invalid integer argument \"%s\" for <%s>", arg, arg_simple->prompt);
    return (FALSE);
  } else {
    if (!(flags&ONLYCHECK)) {
      ADD_VALUE(arg_simple, long, value);
    }
    /* *(long *) arg_simple->store = value; */
    return (TRUE);
  }
}

struct booltab
{
	char	*bname;		/* string to match against */
	char	bneedmatch;	/* number of characters that must match */
	boolean	bval;		/* value to use */
};

static struct booltab	BoolTab[] =
{
	{"yes",		1,	TRUE},
	{"no",		1,	FALSE},
	{"true",		1,	TRUE},
	{"false",	1,	FALSE},
	{"on",		2,	TRUE},
	{"off",		3,	FALSE},
	{NULL,		0,	0}
};

/*ARGSUSED*/
boolean argBoolean
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  struct booltab *b;
  char *cp1;
  int l;
  
  /* convert input to lower case */
  for (l = 0, cp1 = arg; *cp1 != '\0'; l++, cp1++) {
    if (isupper(*cp1))
      *cp1 = tolower(*cp1);
  }
  
  /* search for a match in the table */
  for (b = BoolTab; b->bname != NULL; b++) {
    /* if too short, don't even bother trying */
    if (l < b->bneedmatch)
      continue;
    
    if (bcmp(arg, b->bname, l) == 0) {
      /* got a match */
      if (!(flags&ONLYCHECK)) {
	ADD_VALUE(arg_simple, boolean, b->bval);
      }
      /* *(boolean *) arg_simple->store = b->bval; */
      return (TRUE);
    }
  }
  if (!(flags&(OPTIONAL|ONLYCHECK)))
    ArgError(global, "invalid Boolean argument \"%s\" for <%s>", arg, arg_simple->prompt);
  return (FALSE);
}

/*ARGSUSED*/
boolean argDouble
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  char *argp;
  double value;

  value = strtod(arg, &argp);
  if (*argp != '\0') {
    if (!(flags&(OPTIONAL|ONLYCHECK)))
      ArgError(global, "invalid floating point argument \"%s\" for <%s>",
		  arg, arg_simple->prompt);
    return (FALSE);
  } else {
    if (!(flags&ONLYCHECK)) {
      ADD_VALUE(arg_simple, double, value);
    }
    /* *(double *) arg_simple->store = value; */
    return (TRUE);
  }
}

/*ARGSUSED*/
boolean argReal
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  char *argp;
  double double_value;
  float float_value;

  double_value = strtod(arg, &argp);
  float_value = double_value;
  if (*argp != '\0') {
    if (!(flags&(OPTIONAL|ONLYCHECK)))
      ArgError(global, "invalid floating point argument \"%s\" for <%s>",
		  arg, arg_simple->prompt);
    return (FALSE);
  } else {
    if (!(flags&ONLYCHECK)) {
      if (doubleReals()) {
	ADD_VALUE(arg_simple, double, double_value);
      } else {
	ADD_VALUE(arg_simple, float, float_value);
      }
    }
    /* *(real *) arg_simple->store = value; */
    return (TRUE);
  }
}

/*ARGSUSED*/
boolean argFloat
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  char *argp;
  float value;

  value = strtod(arg, &argp);
  if (*argp != '\0') {
    if (!(flags&(OPTIONAL|ONLYCHECK)))
      ArgError(global, "invalid floating point argument \"%s\" for <%s>",
		  arg, arg_simple->prompt);
    return (FALSE);
  } else {
    if (!(flags&ONLYCHECK)) {
      ADD_VALUE(arg_simple, float, value);
    }
    /* *(float *) arg_simple->store = value; */
    return (TRUE);
  }
}

/*ARGSUSED*/
boolean argFileIn
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  FILE *file;
  int ok = 0;
  int len;
  char *command;
  struct stat statbuf;

  if (flags & ONLYCHECK) {
    if (!strcmp(arg, "-"))
      ok = 1; /* STDIN */
    else if (!strncmp(arg, "<|", 2))
      ok = 2; /* PIPE */
    else if (lstat(arg, &statbuf)==0)
      ok = 3; /* FILE */
    return ok ? TRUE : FALSE;
  } else {
    if (!strcmp(arg, "-"))
      file = stdin;
    else if (!strncmp(arg, "<|", 2))
      file = popen(arg+2, "r");
    else if (lstat(arg, &statbuf)==0) {
      len = strlen(arg);
      if (len>2 && arg[len-1]=='Z' && arg[len-2]=='.') {
	command = CallocOrAbort(char, len+6);
	sprintf(command, "zcat %s", arg);
	file = popen(command, "r");
      } else
	file = fopen(arg, "r");
    }
    if (file==NULL) {
      if (!(flags&(OPTIONAL|ONLYCHECK))) {
	ArgError(global, "cannot open file \"%s\" for <%s> for reading: %s",
		 arg, arg_simple->prompt, strerror(errno));
      }
      return (FALSE);
    } else {
      ADD_VALUE(arg_simple, FILE*, file)
      return (TRUE);
    }
  }
}

/*ARGSUSED*/
boolean argFileOut
  (struct ARG_SIMPLE *arg_simple,
   char *arg,
   int flags,
   struct GLOBAL *global)
{
  FILE *file;

  file = fopen(arg, "w");
  if (file==NULL) {
    if (!(flags&(OPTIONAL|ONLYCHECK))) {
      ArgError(global, "cannot open file \"%s\" for <%s> for writing: %s",
	       arg, arg_simple->prompt, strerror(errno));
    }
    return (FALSE);
  } else {
    if (!(flags&ONLYCHECK))
      ADD_VALUE(arg_simple, FILE*, file)
    else
      fclose(file);
    /* *(int *) arg_simple->store = value; */
    return (TRUE);
  }
}
/* ::::::::::::::
 * argFind.c
 * ::::::::::::::
 */
/* #include "includes.h" */

int plus_minus_options = 1;

struct ARG_DESC *FindOldArgDesc
  (char *name)
{
  if (strcmp(name,"argargs"))
    myname = name;
  return NULL;
}

void StoreArgDesc
  (struct ARG_DESC *ad)
{
}

int GetPlusMinusOptionSwitch(void)
{
  char *p;
  p = getenv("PLUSMINUSOPTIONS");
  if (p==NULL)
    return 0;
  else
    return 1;
}
/* ::::::::::::::
 * argPrint.c
 * ::::::::::::::
 */
/* #include "includes.h" */

/***
 * Functions in this file print structures out as ascii text,
 * controlled by the PRINT_ flags.
 * The ascii text is put in a buffer, and all of these functions
 * assume that buf[-1] is accessible.  When they are called from
 * outside they should be called as follows:
 *
 * char buffer[SIZE];
 * buffer[0] = '\n';
 * ... = SprintArgList(buffer+1, arg_list, flags, indent);
 *
 * i.e. so that buf[-1] is legal and contains '\n'
 */

#define strappendchar(buf, c) (*(buf) = (c), *(++buf) = '\0')

static char *Indent(char *, int, int);
static char *NewLineIndent(char *, int, int);
static char *NewLine(char *, int);
static char *EraseWhiteSpace(char *, int);
static int AdvanceToColumn(int, int, FILE *);
static int help_colon_pos = 30;
static int line_length = 79;
static int tabstop = 8;

void PrintArgDesc
  (struct ARG_DESC *arg_desc,
   int flags,
   int indent,
   FILE *file)
{
  char buffer[10000];
  char *end;
  buffer[0] = '\n';	/* so that indent works properly */
  fprintf(file, "Arguments for \"%s\"\n",arg_desc->name);
  end = SprintArgDesc(buffer+1, arg_desc, flags, indent);
  NewLine(end, flags);
  fputs(buffer+1, file);
}

void PrintArgHelp
  (struct ARG_DESC *arg_desc,
   FILE *file,
   int print_caption)
{
  char buffer[10000];
  char *p;
  int flags = (PRINT_OBJECTS|PRINT_ONEPERLINE|PRINT_HELP
	       |PRINT_TYPES|PRINT_OPTION_SIGN);
  int indent = 1;
  int linepos, after_colon = 0;
  char *next_space;
  buffer[0] = '\n';	/* so that indent works properly */
  if (print_caption)
    fprintf(file, "Arguments and options for \"%s\"\n", arg_desc->name);
  p = SprintArgDesc(buffer+1, arg_desc, flags, indent);
  NewLine(p, flags);
  p = buffer+1;
  linepos = 0;
  while (*p) {
    if (*p==':' && p[-1]==' ' && p[1]==' ') {
      /* want to advance to column <help_colon_pos> */
      linepos = AdvanceToColumn(linepos, help_colon_pos, file);
      after_colon = 1;
    }
    if (*p==' ') {
      /* maybe we want to wrap the line */
      next_space = p+1;
      while (*next_space!='\0' && !isspace(*next_space))
	next_space++;
      if ((next_space-p)+linepos > line_length) {
	fputc('\n', file);
	linepos = 0;
	if (after_colon)
	  linepos = AdvanceToColumn(linepos, help_colon_pos+2, file);
      } else
	fputc(*p, file);
    } else
      fputc(*p, file);
    if (*p=='\n') {
      linepos = 0;
      after_colon = 0;
    } else
      linepos++;
    p++;
  }
}

static int AdvanceToColumn
  (int current,
   int column,
   FILE *file)
{
  int next_tab_stop, i;
  if (current > column) {
    fputc('\n', file);
    current = 0;
  }
  while (current < column) {
    next_tab_stop = (current/tabstop+1)*tabstop;
    if (next_tab_stop<=column) {
      fputc('\t', file);
      current = next_tab_stop;
    } else {
      for (i=current;i<column;i++) {
	fputc(' ', file);
	current++;
      }
    }
  }
  return current;
}

void PrintArgUsage
  (struct ARG_DESC *arg_desc,
   FILE *file,
   int print_usage_label)
{
  char buffer[10000];
  char *p;
  int flags = (PRINT_OBJECTS|PRINT_BRACKETS|PRINT_OPTION_SIGN|PRINT_GROUPPERLINE);
  int indent = 1;
  int linepos = 0;
  char *next_line_start;
  char *next_line_end;
  int next_line_len;
  buffer[0] = '\n';	/* so that indent works properly */
  p = buffer+1;
  if (print_usage_label)
    p = strcpyend(p, "Usage: ");
  else
    p = strcpyend(p, "    ");
  (void) sprintf(p, "%s ", arg_desc->name);
  p += strlen(p);
  p = SprintArgDesc(p, arg_desc, flags, indent);
  NewLine(p, flags);
  p = buffer+1;
  linepos = 0;
  while (*p==' ') {
    fputc(*p, file);
    linepos++;
    p++;
  }
  while (*p) {
    linepos++;
    if (*p=='\n') {
      /* maybe we don't want to put out a newline, instead just a space */
      next_line_start = p+1;
      while (*next_line_start && *next_line_start==' ')
	next_line_start++;
      next_line_end = next_line_start+1;
      while (*next_line_end && *next_line_end!='\n')
	next_line_end++;
      if (*next_line_start)
	next_line_len = next_line_end - next_line_start;
      else
	next_line_len = 0;
      if (next_line_len>0 && (next_line_len + linepos + 10< line_length)) {
	fputc(' ', file);
      } else {
	if (next_line_len>0)
	  fputs("\n      ", file);
	else
	  fputc('\n', file);
	linepos = 0;
      }
    } else
      fputc(*p, file);
    if (*p==' ')
      while (*p && *p==' ')
	p++;
    else
      p++;
  }
}

void PrintArgValues
  (struct ARG_DESC *arg_desc,
   FILE *file)
{
  PrintArgDesc(arg_desc,
	       PRINT_OBJECTS|PRINT_ONEPERLINE|PRINT_TYPES
	       |PRINT_ACTUALS|PRINT_NOT_INDENT|PRINT_ACTUAL_STATUS,
	       1, file);
}

void PrintArgTree
  (struct ARG_DESC *arg_desc,
   FILE *file)
{
  PrintArgDesc(arg_desc, PRINT_OBJECTS|PRINT_BRACKETS|PRINT_STRUCTS
	       |PRINT_ACTUALS|PRINT_OPTION_SIGN|PRINT_SEQS, 0, file);
}

/***
 * WhiteSpace - makes sure there is some whitespace in the
 * buffer
 */
static char *WhiteSpace
  (char *buf)
{
  if (!isspace(buf[-1]))
    strappendchar(buf, ' ');
  return buf;
}

/***
 * Indent(buf, indent, flags)
 * 'indent' spaces are put into buf, if it is positioned at
 * the start of a line (detected by checking if buf[-1]=='\n')
 */
static char *Indent
  (char *buf,
   int indent,
   int flags)
{
  int i;
  if (buf[-1]!='\n')
    return buf;
  else {
    if (!(flags&(PRINT_NOT_INDENT|PRINT_NOT_NEWLINE))) {
      for (i=0;i<(indent*4);i++)
	*buf++ = ' ';
    }
    *buf = 0;
    return buf;
  }
}

char *InitBuf
  (char *buf)
{
  buf[0] = '\n';
  buf[1] = '\0';
  return buf+1;
}

/***
 * NewLineIndent(buf, indent, flags)
 * a NEWLINE + 'indent' spaces are put into buf
 */
static char *NewLineIndent
  (char *buf,
   int indent,
   int flags)
{
  int i;
  buf = EraseWhiteSpace(buf, flags);
  if (!(flags&PRINT_NOT_NEWLINE)) {
    if (buf[-1]!='\n')
      *buf++ = '\n';
    if (!(flags&PRINT_NOT_INDENT)) {
      for (i=0;i<(indent*4);i++)
	*buf++ = ' ';
    }
  } else if (buf[-1]!=' ' && buf[-1]!='\n') {
    *buf++ = ' ';
  }
  *buf = 0;
  return buf;
}

/***
 * NewLineIndent(buf)
 * a NEWLINE is put into buf
 */
static char *NewLine
  (char *buf,
   int flags)
{
  buf = EraseWhiteSpace(buf, flags);
  if (!(flags&PRINT_NOT_NEWLINE)) {
    if (buf[-1]!='\n')
      *buf++ = '\n';
  } else if (buf[-1]!=' ' && buf[-1]!='\n') {
    *buf++ = ' ';
  }
  *buf = 0;
  return buf;
}

/***
 * EraseWhiteSpace(buf, flags)
 * move backwards over white space
 */
static char *EraseWhiteSpace
  (char *buf,
   int flags)
{
  while (buf[-1]==' ')
    buf--;
  *buf = 0;
  return buf;
}

char *SprintArgDesc
  (char *buf,
   struct ARG_DESC *arg_desc,
   int flags,
   int indent)
{
  int i;
  if (flags&PRINT_STRUCTS)
    indent += 1;
  for (i=0;i<arg_desc->n_lists;i++) {
    if (flags&PRINT_STRUCTS) {
      buf = NewLineIndent(buf, indent-1, flags);
      sprintf(buf, "ARG_DESC list %d\n", i);
      buf += strlen(buf);
    }
    buf = SprintArgList(buf, arg_desc->arg_list[i], flags, indent);
    if (flags&PRINT_GROUPPERLINE)
      buf = NewLine(buf, flags);
  }
  return buf;
}

char *SprintArgList
  (char *buf,
   struct ARG_LIST *arg_list,
   int flags,
   int indent)
{
  if (arg_list==NULL)
    return buf;
  if (flags&PRINT_STRUCTS) {
    buf = NewLineIndent(buf, indent, flags);
    sprintf(buf, "ARG_LIST %d parent=%d\n", arg_list->id,
	    (arg_list->parent!=NULL ? arg_list->parent->id : 0));
    buf += strlen(buf);
    indent++;
  }
  buf = SprintArgGroup(buf, arg_list->group, flags, indent);
  if (arg_list->next!=NULL) {
    if (!(flags&INSIDE_GROUP))
      buf = NewLine(buf, flags);
    buf = SprintArgJoin(buf, arg_list->join, flags, indent);
    if (flags&PRINT_STRUCTS)
      indent--;
    buf = SprintArgList(buf, arg_list->next, flags, indent);
  }

  return EraseWhiteSpace(buf, flags);
}

char *SprintArgOption
  (char *buf,
   struct ARG_OPTION *arg_option,
   int flags,
   int indent)
{
  char *pre_indent, *post_indent=NULL;
  if (flags&PRINT_STRUCTS) {
    buf = NewLineIndent(buf, indent, flags);
    buf = strcpyend(buf, "ARG_OPTION\n");
    indent++;
  }
  pre_indent = buf;
  buf = Indent(buf,indent, flags);
  post_indent = buf;
  if (flags&PRINT_OPTION_SIGN) {
    if (arg_option->is_signed)
      if (use_plus_minus_options)
	buf = strcpyend(buf, "+/-");
      else
	buf = strcpyend(buf, "-[no]");
    else
      if (use_plus_minus_options)
	buf = strcpyend(buf, "+");
      else
	buf = strcpyend(buf, "-");
  } else {
    if (flags&PRINT_OPTION_DASH)
      strappendchar(buf, '-');
    if (flags&PRINT_SPECIFIERS) {
      if (arg_option->no_prefix)
	strappendchar(buf, 'F');
      if (arg_option->is_signed)
	strappendchar(buf, 'S');
      else
	strappendchar(buf, 'P');
    }
  }
  if (flags&PRINT_OPTIONS)
    buf = strcpyend(buf, arg_option->name);
  if (flags&PRINT_TYPES) {
    buf = WhiteSpace(buf);
    strappendchar(buf, '(');
    buf = strcpyend(buf, arg_option->is_signed ? "signed" : "presence");
    strappendchar(buf, ')');
  }
  if (flags&PRINT_ACTUALS) {
    strappendchar(buf, '=');
    if (arg_option->is_signed) {
      if (arg_option->actual&OPTION_ON)
	buf = strcpyend(buf, true_string);
      else if (arg_option->actual&OPTION_OFF)
	buf = strcpyend(buf, false_string);
      else {
	buf = strcpyend(buf, "=unset(default=");
	if (*arg_option->sign==TRUE)
	  buf = strcpyend(buf, true_string);
	else if (*arg_option->sign==FALSE)
	  buf = strcpyend(buf, false_string);
	else if (*arg_option->sign==UNDEFINED)
	  buf = strcpyend(buf, undefined_string);
	else
	  buf = strcpyend(buf, bad_boolean_string);
	buf = strcpyend(buf, ")");
      }
    } else {
      if (arg_option->actual&OPTION_PRESENT)
	buf = strcpyend(buf, "present");
      else
	buf = strcpyend(buf, "missing");
    }
    if (arg_option->presence!=NULL) {
      sprintf(buf, " presence=%d", *arg_option->presence);
      buf += strlen(buf);
    }
    if (arg_option->neg_presence!=NULL) {
      sprintf(buf, " neg_presence=%d", *arg_option->neg_presence);
      buf += strlen(buf);
    }
  }
  if (flags&PRINT_OPTIONS)
    strappendchar(buf, ' ');
  if (arg_option->sign!=NULL) {
    if (flags&PRINT_PERCENTS)
      buf = strcatend(buf, "%S ");
    if (flags&PRINT_ADDRESSES) {
      sprintf(buf, "0x%lx ", (long) arg_option->sign);
      buf += strlen(buf);
    }
  }
  if (arg_option->neg_sign!=NULL) {
    if (flags&PRINT_PERCENTS)
      buf = strcatend(buf, "%!S ");
    if (flags&PRINT_ADDRESSES) {
      sprintf(buf, "0x%lx ", (long) arg_option->neg_sign);
      buf += strlen(buf);
    }
  }
  if (arg_option->presence!=NULL) {
    if (flags&PRINT_PERCENTS)
      buf = strcatend(buf, "%P ");
    if (flags&PRINT_ADDRESSES) {
      sprintf(buf, "0x%lx ", (long) arg_option->presence);
      buf += strlen(buf);
    }
  }
  if (arg_option->neg_presence!=NULL) {
    if (flags&PRINT_PERCENTS)
      buf = strcatend(buf, "%!P ");
    if (flags&PRINT_ADDRESSES) {
      sprintf(buf, "0x%lx ", (long) arg_option->neg_presence);
      buf += strlen(buf);
    }
  }
  if (arg_option->help!=NULL) {
    if (flags&PRINT_PERCENTS)
      buf = strcatend(buf, "%? ");
    if (flags&PRINT_ADDRESSES) {
      sprintf(buf+strlen(buf), "0x%lx ", (long) arg_option->help);
      buf += strlen(buf);
    }
    if (flags&PRINT_HELP) {
      sprintf(buf, ": %s\n", arg_option->help);
      buf += strlen(buf);
    }
  }

  if (buf==post_indent) {
    buf = pre_indent;
    *buf = '\0';
  } else {
    if (flags&PRINT_ONEPERLINE)
      buf = NewLine(buf, flags);
  }
  
  return EraseWhiteSpace(buf, flags);
}

char *SprintArgGroup
  (char *buf,
   struct ARG_GROUP *arg_group,
   int flags,
   int indent)
{
  char *pre_indent, *post_indent=NULL;
  if (flags&PRINT_STRUCTS) {
    buf = NewLineIndent(buf, indent, flags);
    buf = strcpyend(buf, "ARG_GROUP");
    if ((flags&PRINT_ACTUALS) && arg_group->presence!=NULL) {
      sprintf(buf," presence=%d", *arg_group->presence);
      buf += strlen(buf);
    }
    if ((flags&PRINT_ACTUALS) && arg_group->neg_presence!=NULL) {
      sprintf(buf," neg_presence=%d", *arg_group->neg_presence);
      buf += strlen(buf);
    }
    strappendchar(buf, '\n');
    indent++;
  }
  pre_indent = buf;
  if (flags&PRINT_PARENS) {
    buf = Indent(buf, indent, flags);
    buf = strcpyend(buf, "( ");
  }
  if (arg_group->tag==ARG_GROUP_SIMPLE) {
    buf = SprintArgSimple(buf, arg_group->u.simple.arg_simple, flags, indent);
  } else if (arg_group->tag==ARG_GROUP_OPTION) {
    if (arg_group->u.option.optional && (flags&PRINT_BRACKETS)) {
      buf = Indent(buf, indent, flags);
      buf = WhiteSpace(buf);
      strappendchar(buf, '[');
    }
    if (arg_group->u.option.option!=NULL)
      buf = SprintArgOption(buf, arg_group->u.option.option, flags, indent);
    if (arg_group->u.option.arg_list!=NULL)
      buf = SprintArgList(buf, arg_group->u.option.arg_list,flags|INSIDE_GROUP,
		    indent+(1-arg_group->u.option.optional));
    if (arg_group->u.option.optional && (flags&PRINT_BRACKETS)) {
      /* buf = WhiteSpace(buf); */
      strappendchar(buf, ']');
    }
  }

  if (arg_group->presence!=NULL) {
    if (flags&PRINT_PERCENTS)
      buf = strcatend(WhiteSpace(buf), "%P");
    if (flags&PRINT_ADDRESSES) {
      sprintf(WhiteSpace(buf), "0x%lx", (long) arg_group->presence);
      buf += strlen(buf);
    }
    if (flags&PRINT_ACTUALS) {
      buf = strcatend(WhiteSpace(buf), "presence=");
      buf = strcatend(buf, BooleanString(*arg_group->presence));
    }
  }
  if (arg_group->neg_presence!=NULL) {
    if (flags&PRINT_PERCENTS)
      buf = strcatend(WhiteSpace(buf), "%!P");
    if (flags&PRINT_ADDRESSES) {
      sprintf(WhiteSpace(buf), "0x%lx", (long) arg_group->neg_presence);
      buf += strlen(buf);
    }
    if (flags&PRINT_ACTUALS) {
      buf = strcatend(WhiteSpace(buf), "neg_presence=");
      buf = strcatend(buf, BooleanString(*arg_group->neg_presence));
    }
  }

  if (arg_group->help!=NULL) {
    if (flags&PRINT_PERCENTS)
      buf = strcatend(WhiteSpace(buf), "%?");
    if (flags&PRINT_ADDRESSES) {
      sprintf(WhiteSpace(buf), "0x%lx", (long) arg_group->help);
      buf += strlen(buf);
    }
    if (flags&PRINT_HELP) {
      sprintf(WhiteSpace(buf), ": %s", arg_group->help);
      buf += strlen(buf);
    }
  }

  if (flags&PRINT_PARENS)
    buf = strcatend(WhiteSpace(buf), ")");

  if ((flags&PRINT_ONEPERLINE) && arg_group->tag==ARG_GROUP_SIMPLE)
    buf = NewLine(buf, flags);

  if (buf==post_indent) {
    buf = pre_indent;
    *buf = '\0';
  }

  return EraseWhiteSpace(buf, flags);
}

char *SprintArgSimple
  (char *buf,
   struct ARG_SIMPLE *arg_simple,
   int flags,
   int indent)
{
  char *pre_indent,*post_indent;
  if (flags&PRINT_STRUCTS) {
    buf = NewLineIndent(buf, indent, flags);
    (void) sprintf(buf, "ARG_SIMPLE %d\n", arg_simple->id);
    buf += strlen(buf);
    indent++;
  }
  pre_indent = buf;
  buf = Indent(buf, indent, flags);
  post_indent = buf;
  if (arg_simple->type==ARG_LITERAL) {
    if (flags&PRINT_LITERALS) {
      sprintf(WhiteSpace(buf), "'%s'", arg_simple->prompt);
      buf += strlen(buf);
    }
  } else {
    if (flags&PRINT_SIMPLES) {
      if (buf[-1]!='[')
	buf = WhiteSpace(buf);
      sprintf(buf, "<%s>", arg_simple->prompt);
      buf += strlen(buf);
      if (arg_simple->vector)
	buf = strcatend(buf, "...");
    }
    if (flags&PRINT_TYPES) {
      buf = WhiteSpace(buf);
      strappendchar(buf, '(');
      buf = strcpyend(buf, arg_type_name[(int) arg_simple->type]);
      strappendchar(buf, ')');
    }
    if (flags&PRINT_SPECIFIERS) {
      strappendchar(buf, '%');
      if (arg_simple->verify!=NULL)
	strappendchar(buf, 'V');
      if (arg_type_char[(int)arg_simple->type]=='\0')
	ErrorAbort("bad type specifier %d", arg_simple->type);
      else
	strappendchar(buf, arg_type_char[(int) arg_simple->type]);
    }
    if (arg_simple->verify && (flags&PRINT_ADDRESSES)) {
      sprintf(WhiteSpace(buf), "0x%lx", (long) arg_simple->verify);
      buf += strlen(buf);
    }
    if (flags&PRINT_ADDRESSES) {
      sprintf(WhiteSpace(buf), "0x%lx", (long) arg_simple->store);
      buf += strlen(buf);
    }
    if (flags&PRINT_ACTUALS) {
      buf = strcpyend(buf, " = ");
      if (arg_simple->vector) {
	void **ptr;
	strappendchar(buf, '{');
	if (arg_simple->store!=NULL && *(void***)arg_simple->store!=NULL)
	  for (ptr = *(void***)arg_simple->store; *ptr!=NULL; ptr++) {
	    buf = SprintArgValue(buf, arg_simple->type, (void*)ptr, flags);
	    if (ptr[1]!=NULL)
	      strappendchar(buf, ' ');
	  }
	strappendchar(buf, '}');
	if (arg_simple->store_len) {
	  sprintf(buf, " (len=%d)", *arg_simple->store_len);
	  buf += strlen(buf);
	}
      } else {
	buf = SprintArgValue(buf, arg_simple->type, arg_simple->store, flags);
      }
    }
    if (flags&PRINT_ACTUAL_STATUS) {
      if (arg_simple->actual&FOUND_ACTUAL) {
	buf = strcpyend(WhiteSpace(buf), "(found value)");
      } else if (arg_simple->actual&ACTUAL_CLEARED) {
	buf = strcpyend(WhiteSpace(buf), "(found value but list cleared)");
      } else {
	buf = strcpyend(WhiteSpace(buf), "(original value)");
      }
    }
  }
  if (buf==post_indent) {
    buf = pre_indent;
    *buf = '\0';
  }
  return EraseWhiteSpace(buf, flags);
}

char *SprintArgValue
  (char *buf,
   enum ARG_TYPE type,
   void *ptr,
   int flags)
{
  if (type==ARG_REAL) type = doubleReals() ? ARG_DOUBLE : ARG_FLOAT;
  switch (type) {
  case ARG_FLOAT:
    (void) sprintf(buf, "%g", *(float *)ptr);
    break;
  /* case ARG_REAL:
   *   (void) sprintf(buf, "%g", *(real *)ptr);
   *   break;
   */
  case ARG_DOUBLE:
    (void) sprintf(buf, "%g", *(double *)ptr);
    break;
  case ARG_BOOLEAN:
    (void) sprintf(buf, "%s", BooleanString(*(boolean *)ptr));
    break;
  case ARG_INTEGER:
  case ARG_NATURAL:
    (void) sprintf(buf, "%d", *(int *)ptr);
    break;
  case ARG_SHORT:
    (void) sprintf(buf, "%d", *(short *)ptr);
    break;
  case ARG_LONG:
    (void) sprintf(buf, "%ld", *(long *)ptr);
    break;
  case ARG_CHAR:
    if (isprint(*(char *)ptr))
      (void) sprintf(buf, "'%c'", *(char *)ptr);
    else
      (void) sprintf(buf, "'\\%03o'", *(char *)ptr);
    break;
  case ARG_STRING:
  case ARG_FORMAT:
  case ARG_AREA_FORMAT:
  case ARG_RANGE_STRING:
  case ARG_VAR_NAME:
  case ARG_MEMBER_NAME:
    (void) sprintf(buf, "\"%s\"", *(char **)ptr);
    break;
  case ARG_FILE_IN:
    (void) sprintf(buf, "0x%lx", (long) *(FILE **)ptr);
    break;
  case ARG_FILE_OUT:
    (void) sprintf(buf, "0x%lx", (long) *(FILE **)ptr);
    break;
  default:
    assert(0);
  }
  return buf + strlen(buf);
}

char *SprintArgJoin
  (char *buf,
   enum ARG_JOIN join,
   int flags,
   int indent)
{
  char *pre_indent,*post_indent;
  pre_indent = buf;
  buf = Indent(buf, indent, flags);
  buf = WhiteSpace(buf);
  post_indent = buf;
  switch (join) {
  case ARG_OR:
    buf = strcpyend(buf, "|");
    break;
  case ARG_AND:
    buf = strcpyend(buf, "&");
    break;
  case ARG_SEQ:
    if (flags&PRINT_SEQS)
      buf = strcpyend(buf, "+");
    break;
  case ARG_XOR:
    buf = strcpyend(buf, "^");
    break;
  default:
    break;
  }
  if (buf==post_indent) {
    buf = pre_indent;
    *buf = '\0';
  }
  return EraseWhiteSpace(buf, flags);
}

char *FlagPrefix
  (int flag)
{
  if (flag) {
    if (use_plus_minus_options)
      return "+";
    else
      return "-";
  } else {
    if (use_plus_minus_options)
      return "-";
    else
      return "-no";
  }
}
/* ::::::::::::::
 * miscStr.c
 * ::::::::::::::
 */
/* #include "includes.h" */

char *pchar(char c)
{
  static char buf[10];
  if (isprint(c)) {
    buf[0] = buf[2] = '\'';
    buf[1] = c;
  } else if (c=='\0')
    strcpy(buf,"(nullchar)");
  else
    sprintf(buf,"(char:%d)",c);
  return buf;
}

void strncpy_null(char *dest, char *src, int n)
{
  strncpy(dest,src,n-1);
  dest[n-1] = '\0';
}

/***
 * copies str to buf and returns a pointer to the end of the string
 */
char *strcpyend(char *buf, char *str)
{
  while (*str) *(buf++) = *(str++);
  *buf = 0;
  return buf;
}

/***
 * concats str to buf and returns a pointer to the end of the string
 */
char *strcatend(char *buf, char *str)
{
  while (*buf) buf++;
  while (*str) *(buf++) = *(str++);
  *buf = 0;
  return buf;
}

/* ::::::::::::::
 * sizeOfReal.c
 * ::::::::::::::
 */
/* #include <include/utils.h> */

static int double_reals = -1;
int sizeOfReal;

void ArgsSetRealSize(int size)
{
  if (size==sizeof(float))
    double_reals = 0;
  else if (size==sizeof(double))
    double_reals = 1;
  else
    ErrorAbort("ArgsSetRealSize: illegal size (%d) (expected %d or %d)",
	       size, sizeof(float), sizeof(double));
}

int doubleReals()
{
  if (double_reals==-1) {
    if (sizeOfReal==sizeof(float))
      double_reals = 0;
    else if (sizeOfReal==sizeof(double))
      double_reals = 1;
    else
      ErrorAbort("Args: must set size of reals in main program with statement \"ArgsSetRealSize(sizeof(real));\"");
  }
  return double_reals;
}
