/* Infix notation expression parser--expr */

%{
#include <stdio.h>
#include <stdlib.h>
#ifdef USE_TAP_LIBRARIES
  #define NO_String_TYPE
  #include <tap/argf.h>
  #include <tap/tapmath.h>
  #include <tap/taputil.h>
  #include <tap/HashTable.h>
#else
  #include "config.h"
  #include "real.h"
  #include "hashTable.h"
#endif /* USE_TAP_LIBRARIES */
#include "vcalc.h"
struct ExprNode *yyroot;
int yyparse(void);
void yyerror(char *);
int yylex(void);
/* couple of things to stop compiler warnings on bison code */
/* no longer seem necessary (Apr/98) */
/* static void __yy_bcopy(char *, char *, int); */
/* void * alloca(unsigned int); */
%}

%union {
  real			val;
  char*			str;
  struct ExprNode	*node;
}

%token <val> RNUM
%token <val> INUM
%token <str> IDENT
%token <str> STRING
%token <str> NEWLINE
%token <str> COLON
%token <str> SEMICOLON
%token <str> LPAREN
%token <str> RPAREN
%token <str> LANGLE
%token <str> RANGLE
%token <str> LSQUARE
%token <str> RSQUARE
%token <str> LCURLY
%token <str> RCURLY
%token <str> BAR
%token <str> BLANK
%right <str> COMMA
%right <str> EQUALS DBLEQUALS NOTEQUALS LESSEQUAL GRTREQUAL LESS GRTR
%left <str> DOT
%left <str> MINUS PLUS
%right <str> HASH
%left <str> DIVIDE
%left <str> TIMES
%left <str> CDIVIDE
%left <str> CTIMES
%left <str> TTIMES
%left <str> NEG     /* negation--unary minus */
%left <str> BANG    /* boolean not */
%right <str> EXP    /* exponentiation        */
%right <str> HAT    /* exponentiation        */
%right <str> QUOTE

%type <node> input end expr list sequence vector locals

/* Grammar follows */
%%

input:
expr end		{ $$ = yyroot = $1; YYACCEPT; }
| STRING expr end	{ $$ = yyroot = newNode(Label, 0, $2, NULL);
			  $$->name = $1;
			  YYACCEPT; }
|			{ yyroot = NULL; YYACCEPT; }
;

end:
NEWLINE			{}
| SEMICOLON		{}
;

list:
  expr			{ $$ = newNode(List, COMMA, $1, NULL); }
| expr COMMA list	{ $$ = newNode(List, COMMA, $1, $3); }
;

locals:
			{ $$ = NULL; }
| SEMICOLON list	{ $$ = $2; }
;

sequence:
  expr			{ $$ = $1; }
| expr COMMA sequence	{ $$ = newNode(Sequence, 0, $1, $3); }
;

vector:
  			{ $$ = NULL; }
| COMMA expr vector	{ $$ = newNode(List, BLANK, $2, $3); }
| expr vector		{ $$ = newNode(List, BLANK, $1, $2); }
;

expr:	INUM		{ $$ = newNode(Literal, 0, NULL, NULL);
			  $$->var->type = Scalar;
			  $$->var->d.value = $1; }
|	RNUM		{ $$ = newNode(Literal, 0, NULL, NULL);
			  $$->var->type = Scalar;
			  $$->var->d.value = $1; }
|	STRING		{ $$ = newNode(Literal, 0, NULL, NULL);
			  $$->var->type = String;
			  $$->var->d.str = $1; }
|	IDENT		{ $$ = newNode(Variable, 0, NULL, NULL);
			  $$->name = $1; }
| IDENT LPAREN list locals RPAREN {
			  $$ = newNode(Function, 0, $3, $4);
			  $$->name = $1; }
| IDENT LPAREN RPAREN	{ $$ = newNode(Function, 0, NULL, NULL);
			  $$->name = $1; }
| IDENT LPAREN locals RPAREN	{ $$ = newNode(Function, 0, NULL, $3);
			  $$->name = $1; }
| LANGLE list locals RANGLE	{ $$ = newNode(Operator, LANGLE, $2, $3); }
| expr PLUS expr	{ $$ = newNode(Operator, PLUS, $1, $3); }
| expr MINUS expr	{ $$ = newNode(Operator, MINUS, $1, $3); }
| expr EQUALS expr	{ $$ = newNode(Operator, EQUALS, $1, $3); }
| expr DBLEQUALS expr	{ $$ = newNode(Operator, DBLEQUALS, $1, $3); }
| expr NOTEQUALS expr	{ $$ = newNode(Operator, NOTEQUALS, $1, $3); }
| expr TIMES expr	{ $$ = newNode(Operator, TIMES, $1, $3); }
| expr CTIMES expr	{ $$ = newNode(Operator, TIMES, $1, $3); }
| expr HASH expr	{ $$ = newNode(Operator, HASH, $1, $3); }
| expr DIVIDE expr     	{ $$ = newNode(Operator, DIVIDE, $1, $3); }
| expr CDIVIDE expr    	{ $$ = newNode(Operator, DIVIDE, $1, $3); }
| expr DOT expr     	{ $$ = newNode(Operator, DOT, $1, $3); }
| MINUS expr %prec NEG	{ $$ = newNode(Operator, NEG, $2, NULL); }
| BANG expr		{ $$ = newNode(Operator, BANG, $2, NULL); }
| expr QUOTE		{ $$ = newNode(Operator, QUOTE, $1, NULL); }
| expr EXP expr		{ $$ = newNode(Operator, EXP, $1, $3); }
| expr HAT expr		{ $$ = newNode(Operator, HAT, $1, $3); }
| expr LESS expr	{ $$ = newNode(Operator, LESS, $1, $3); }
| expr GRTR expr	{ $$ = newNode(Operator, GRTR, $1, $3); }
| expr LESSEQUAL expr	{ $$ = newNode(Operator, LESSEQUAL, $1, $3); }
| expr GRTREQUAL expr	{ $$ = newNode(Operator, GRTREQUAL, $1, $3); }
| LPAREN sequence RPAREN{ $$ = newNode(Operator, LPAREN, $2, NULL); }
| LSQUARE expr RSQUARE	{ $$ = newNode(Operator, LSQUARE, $2, NULL); }
| LCURLY vector RCURLY	{ $$ = newNode(Operator, LCURLY, $2, NULL); }
| BAR expr BAR		{ $$ = newNode(Operator, BAR, $2, NULL); }
;

%%

#include <stdio.h>

void yyerror (s)  /* Called by yyparse on error */
     char *s;
{
  while (yylex() != NEWLINE) ;
  AbortExpr("parser: %s", s);
}

struct ExprNode *ParseExpr()
{
  /* extern int yydebug; */
  /* yydebug = 1; */
  inputLine[0] = '\0';
  inputLineEnd = inputLine;
  yyparse ();
  return yyroot;
}
