
/*  copyright 1990  Richard H. Krukar all rights reserved

	Permission granted to buy, sell, or steal this software is granted.
    The author retains the right to distribute this software freely, but
    is not responsible for it's quality or maintainance. */

#include "fft.h"

void fft
  (struct complex *x,
   int length)
{
    dintime(x, length, wtab );
    bitrev((real *) x, length );
}


void ifft
  (struct complex *x,
   int length)
{
    real scale;
    int i;

    idintime(x, length, wtab );
    bitrev((real *) x, length );

    
    scale = (1.0/length);
    for(i=0; i<length; i++) {
      x[i].r *= scale;
      x[i].i *= scale;
    }
}

void rfft
  (real *x_in,
   int length)
{
    int hlen, step, i, c, s;
    struct complex odd, even;
    struct complex *x = (struct complex *)x_in;

    hlen = length/2;

    dintime( x, hlen, wtab );
    bitrev( (real *)x, hlen );

    c = step = TLEN/length;
    s = c + TLEN/4;

    for(i=1; i<=(length/4); i++) {
	even.r = x[i].r + x[hlen-i].r;
	even.i = x[i].i - x[hlen-i].i;

	odd.r = x[i].r - x[hlen-i].r;
	odd.i = x[i].i + x[hlen-i].i;

	x[i].i = odd.i * wtab[s] - odd.r * wtab[c]; /* use x as a tmp */
	odd.r = odd.r * wtab[s] + odd.i * wtab[c];
	odd.i = x[i].i;

	x[i].r = .5*(even.r+odd.r);
	x[i].i = .5*(even.i+odd.i);
	x[hlen-i].r = .5*(even.r-odd.r);
	x[hlen-i].i = .5*(odd.i-even.i);
	c += step;
	s += step;
	if( s > TLEN ) s -= TLEN;
    }
    even.r = x[0].r + x[0].i;
    x[0].i = x[0].r - x[0].i;
    x[0].r = even.r;
}


void rifft
  (real *x_in,
   int length)
{
    int hlen, step, i, c, s;
    struct complex odd, even;
    real *fptr, *eptr, scale;
    struct complex *x = (struct complex *)x_in;

    hlen = length/2;
    c = step = TLEN/length;
    s = c + 3*TLEN/4;

    for(i=1; i<=(length/4); i++) {
	even.r = x[i].r + x[hlen-i].r;
	even.i = x[i].i - x[hlen-i].i;

	odd.r = x[i].r - x[hlen-i].r;
	odd.i = x[i].i + x[hlen-i].i;

	x[i].i = odd.i * wtab[s] - odd.r * wtab[c]; /* use x as a tmp */
	odd.r = odd.r * wtab[s] + odd.i * wtab[c];
	odd.i = x[i].i;

	x[i].r = .5*(even.r-odd.r);
	x[i].i = .5*(even.i-odd.i);
	x[hlen-i].r = .5*(even.r+odd.r);
	x[hlen-i].i = .5*(-odd.i-even.i);
	c += step;
	s += step;
	if( s > TLEN ) s -= TLEN;
    }
    even.r = .5*(x[0].r + x[0].i);
    x[0].i = .5*(x[0].r - x[0].i);
    x[0].r = even.r;
    idintime( x, hlen, wtab );
    bitrev( (real *)x, hlen );

    scale = (1.0/hlen); eptr = (real *)&x[length>>1];
    for(fptr = (real *)&x[0]; fptr<eptr; fptr++)
	*fptr *= scale;
}
