#ifndef _HashTable_
#define _HashTable_

#ifndef _PFI_
#define _PFI_
typedef int (*PFI)(); /* pointer to function returning int */
#endif

struct BINDING;	/* `forward' declaration of BINDING */

typedef struct HashTableItem {
  int			context;
  int			id;	/* begins numbering at 1 */
  union {
    void		*ptr;	/* to an arbitrary structure or function */
    long		token;  /* or a token identifier */
    int			count;
    real		value;
    PFI			pfi;
  } data;
  char			*name;
  struct HashTableItem	*next;	/* in this hash slot */
} HashTableItem;

typedef void *(*HashTableDataDeAllocater)(void *);

typedef struct HashTable {
  HashTableItem	**hash_vec;
  HashTableItem	**id_vec;
  HashTableItem	**alpha_vec; /* unimplemented */
  int		hash_bins;
  int		max_items;
  int		free_id;
  int		free_context;
  HashTableDataDeAllocater	free_item;
} HashTable;

#define HT_NOT_FOUND	1
#define HT_AMBIGUOUS	2
#define HT_DUPLICATE	3
#define HT_BAD_ID	4
#define HT_NO_MEMORY	5
#define HT_BAD_NAME	6
#define HT_OVERFLOW	7
#define HT_CANT_DO	8

extern int hash_errno;
extern char *hash_error[];

HashTable *HashTableNewTable(HashTable *, int max_items, int hash_bins, int index_types);
		/* (ht, max_items, hash_bins, index_types) */
/* index_types can be one of the following: */
#define ID_INDEX	1
#define ALPHA_INDEX	2
#define HashTableDeAllocator(HashTable, DeAllocater) {(HashTable)->free_item = (DeAllocater);}
HashTableItem *HashTableLookup(HashTable *, char *name, int context);
int	  HashTableGetId(HashTable *, char *name, int context);
int	  HashTableNewContext(HashTable *);
HashTableItem *HashTableLookupById(HashTable *, int id);
HashTableItem *HashTableInsert(HashTable *, char *name, int context, void *data);
HashTableItem *HashTableNext(HashTable *, int *id);
		/* (ht, &id) - id is a reference, start with it at 0 */
HashTableItem *HashTableNextByContext(HashTable *, int *id, int context);
		/* (ht, &id, context) - id is a reference, start with it at 0 */
#define HashTableFirstByContext(ht, id, context) (*id=0, HashTableNextByContext(ht, id, context))
int	  HashTableDelete(HashTable *, char *name, int context);
int	  HashTableDeleteByContext(HashTable *, int context);

#endif /* _HashTable_ */
