#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include "utils.h"
#include "config.h"

FILE *dout; /* = stdout; initialized in main() was (&__iob[1]); */
FILE *din; /* = stdin; initialized in main() was (&__iob[0]); */

char *myname;
char *true_string = "true";
char *false_string = "false";
char *undefined_string = "undefined";
char *bad_boolean_string = "bad_boolean";

int stand_alone = 1;

int IsAsciiString
  (char *p)
{
  if (! *p)
    return 0;
  for (; *p; p++)
    if (!isascii(*p))
      return 0;
  return 1;
}

char *BooleanString
  (int b)
{
  switch (b) {
  case 0: return false_string;
  case -1: return undefined_string;
  case 1: return true_string;
  default: return bad_boolean_string;
  }
}

int IsLegalVarName
  (char *name)
{
  if (!name || !*name) return 0;
  if (!(   (*name >= 'a' && *name <= 'z')
	|| (*name >= 'A' && *name <= 'Z')
	|| *name=='$' || *name=='@')) return 0;
  while (*name)
    if (   (*name >= 'a' && *name <= 'z')
	|| (*name >= 'A' && *name <= 'Z')
	|| (*name >= '0' && *name <= '9')
	|| *name=='_' || *name=='-' || *name=='$' || *name=='@') name++;
    else
      return 0;
  return 1;
}

void *callocOrAbort(unsigned int nelem, unsigned int elsize)
{
  char *p;
  if (nelem==0 || elsize==0)
    return NULL;
  p = calloc(nelem, elsize);
  if (p==NULL)
    ErrorAbort("callocOrAbort: out of memory (%d,%d)", nelem, elsize);
  return p;
}

void *reallocOrAbort(void *old, unsigned int size)
{
  char *p;
  if (size==0) {
    if (old!=NULL)
      free(old);
    return NULL;
  }
  if (old==NULL)
    p = calloc(size, 1);
  else
    p = realloc(old, size);
  if (p==NULL)
    ErrorAbort("reallocOrAbort: out of memory (%d)", size);
  return p;
}

char *SysError
  (int n)
{
  return(strerror(n));
  /*
   * old, deprecated code
  static char str[40];
  if (n>0 || n<sys_nerr) return sys_errlist[n];
  else {
    sprintf(str, "No error message for error #%d", n);
    return str;
  }
  */
}

void
  Abort (void)
{
  exit(1);
}

/***
 * ErrorV takes a pointer to a va_list
 */
void ErrorV(char *format, va_list *args)
{
  
  (void) fprintf(stderr, "%s: ", myname);
  (void) vfprintf(stderr, format, *args);
  va_end(*args);
  (void) fputc('\n', stderr);
}

/*  Error should be called like:
 *    Error(format, arg1, arg2...);
 */

void
  Error (char *format, ...)
{
  va_list args;

  (void) fprintf(stderr, "%s: ", myname);
  va_start(args, format);
  (void) vfprintf(stderr, format, args);
  va_end(args);
  (void) fputc('\n', stderr);
}

void
  ErrorAbort (char *format, ...)
{
  va_list args;
  
  (void) fprintf(stderr, "%s: ", myname);
  va_start(args, format);
  (void) vfprintf(stderr, format, args);
  va_end(args);
  (void) fputc('\n', stderr);
  exit(1);
}

